﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace WebMetar_NS 
{
    /// <summary>
    /// Colored tooltip - uses RichText box which supports multi-colored text.
    ///     1. Auto increase vertical size to fit text.
    ///     2. Fad in/out on timer.
    ///     3. Click on this dialog and timer stops, right click to drag UI to new position.
    /// 
    /// Author: Dennis Lang 2009
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    /// </summary>
    public partial class ColorToolTip : Form
    {
        public ColorToolTip()
        {
            InitializeComponent();

            // init tag to junk value
            this.Tag = -1;  

            // default to invisible
            this.Opacity = 0;

            // calculate row height, 1.2 gives some space around text.
            Graphics g = this.CreateGraphics();
            rowHeight = (int)(g.MeasureString("Ag", this.richText.Font).Height * 1.2);
        }

        int rowHeight;
        double fadeDir;
        uint millsecBeforeOff = 3000;
        Stopwatch stopwatch = new Stopwatch();

        /// <summary>
        /// Set/get milliseconds to start fade off.
        /// </summary>
        public uint Milliseconds
        {
            set { millsecBeforeOff = value; }
            get { return millsecBeforeOff; }
        }

        /// <summary>
        /// Set text (no special coloriziation).
        /// </summary>
        public string Message
        {
            set { this.richText.Text = value; }
            get { return this.richText.Text; }
        }

        public void ResetClock()
        {
            timer.Enabled = true;
            stopwatch.Reset();
            stopwatch.Start();
            fadeDir = +0.20;
        }

        /// <summary>
        /// Clear text
        /// </summary>
        public void Clear()
        {
            this.richText.Clear();
            this.Opacity = 0;
            this.Visible = false;
            this.stopwatch.Stop();
        }

        /// <summary>
        /// Append text with optional color and option font.
        /// </summary>
        /// <param name="text"></param>
        /// <param name="color"></param>
        /// <param name="font"></param>
        public void Append(string text, Color color, Font font)
        {
            if (font != null)
            {
                richText.SelectionFont = font;
                Graphics g = this.CreateGraphics();
                rowHeight = (int)(g.MeasureString("Ag", font).Height * 1.2);
            }

            if (color != null)
                richText.SelectionColor = color;

            richText.SelectedText = text;
        }

        /// <summary>
        /// Access richtext box to set color and font on selected text.
        /// </summary>
        public RichTextBox RichText
        {
            get { return this.richText; }
        }

        protected override void OnVisibleChanged(EventArgs e)
        {
            base.OnVisibleChanged(e);
            if (this.Visible)
            {
                this.FormBorderStyle = FormBorderStyle.None;
                this.Height = rowHeight;
                ResetClock();
                CheckSize();
            }
            else
            {
                this.timer.Enabled = false;
            }
        }

        /// <summary>
        /// Auto size height to prevent veritcal scrollbar from appearing.
        /// </summary>
        bool lastVsVisible = false;
        private bool CheckSize()
        {
            if (fadeDir > 0)
            {
                bool vsVisible = HasVerticalScrollBar();

                vsVisible = vsVisible && (this.Height < rowHeight * 20);
                if (vsVisible)
                {
                    this.Height += rowHeight;
                }
                lastVsVisible = vsVisible;
            }

            return lastVsVisible;
        }

        private bool HasVerticalScrollBar()
        {
            Control ctl = richText;
            int wndStyle = Win32.GetWindowLong(ctl.Handle, Win32.GWL_STYLE);
            return (wndStyle & Win32.WS_VSCROLL) != 0;
        }

        public class Win32
        {
            // offset of window style value
            public const int GWL_STYLE = -16;

            // window style constants for scrollbars
            public const int WS_VSCROLL = 0x00200000;
            public const int WS_HSCROLL = 0x00100000;

            [DllImport("user32.dll", SetLastError = true)]
            public static extern int GetWindowLong(IntPtr hWnd, int nIndex);
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            this.Opacity = Math.Max(0.0, Math.Min(1.0, this.Opacity + fadeDir));
            this.Visible = (this.Opacity != 0);
            if (stopwatch.ElapsedMilliseconds > this.millsecBeforeOff)
                fadeDir = -0.05;
            else
                CheckSize();
        }

        /// <summary>
        /// Place mouse over tip viewer to prevent it from fading.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void richText_MouseEnter(object sender, EventArgs e)
        {
            this.Opacity = 1.0;
            this.fadeDir = 0;
            this.stopwatch.Stop();
            this.timer.Enabled = CheckSize();
            // this.closeBtn.Visible = true;
            this.FormBorderStyle = FormBorderStyle.Sizable;
        }

        private void richText_MouseLeave(object sender, EventArgs e)
        {
            Point mousePos = System.Windows.Forms.Control.MousePosition;

            if (false == HasVerticalScrollBar() || mousePos.X < this.Width)
            {
                // this.fadeDir = -0.1;
                // this.timer.Enabled = true;
                // this.stopwatch.Start();

                // this.Close();
            }
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            this.Visible = false;
            this.stopwatch.Stop();
            this.timer.Enabled = false;
            e.Cancel = true;
            base.OnClosing(e);
        }

        /// <summary>
        /// Press Right mouse to move tip viewer.
        /// </summary>
        Point startDrag = Point.Empty;
        private void richText_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                Point mousePos = System.Windows.Forms.Control.MousePosition;
                if (startDrag != Point.Empty)
                {
                    Size move = new Size(mousePos) - new Size(startDrag);
                    this.Location += move;
                }
                startDrag = mousePos;
            }
            else
            {
                startDrag = Point.Empty;
            }
        }

        private void richText_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.Control)
            {
                if (this.richText.SelectedText.Length != 0) 
                {
                    switch (e.KeyCode)
                    {
                        case Keys.C:    // copy selected text to clip board.
                            Clipboard.SetText(this.richText.SelectedText);
                            break;
                        case Keys.V:
                            // Clipboard.GetText();
                            break;
                    }

                }
            }
        }

    }
}
