﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

using System.Net;
using System.Web;
using System.IO;
using System.Threading;
using System.Xml;
using System.Xml.XPath;
using System.Text.RegularExpressions;
using System.Collections;

using System.Net.NetworkInformation;
using System.Reflection;
using System.IO.Compression;
using System.Drawing.Drawing2D;


namespace WebMetar_NS
{
    /// <summary>
    /// Web METAR fetch and display program.
    /// * Define URL's to pull station data from (raw or text)
    /// * Define METAR table column order, units and formatting.
    /// * Master and User station list editor
    /// * Timer fetches data and updates table with any changes.
    /// 
    /// Credits:
    ///   METAR decoder based off of work by rantanplanisback http://www.codeproject.com/KB/cs/weathernotify.aspx
    ///   
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// TODO - 
    ///     1. Add support for icons for Cloud/Sky and Weather conditions
    ///     2. Add support for day/night icons
    ///     3. Add XML decoder
    ///     
    /// </summary>
    public partial class MainForm : Form
    {
        public MainForm()
        {
            InitializeComponent();

            InitializeWebMetar();
        }

        private void InitializeWebMetar()
        {
            // Setup column sorting
            stationListView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAlpha);
            webResponseView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAlpha);
            tableDataView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAlpha);

            ListViewHelper.EnableDoubleBuffer(stationListView);
            ListViewHelper.EnableDoubleBuffer(webResponseView);
            ListViewHelper.EnableDoubleBuffer(tableDataView);

            // Initialize graphs with default settings and wire up properties
            this.barGraph.DefaultSettings();
            this.barGraphProperty.SelectedObject = this.barGraph.Settings;

            this.timeGraph.DefaultSettings();
            this.timeGraphProperty.SelectedObject = this.timeGraph.Settings;

            // Set Detail zone to full Bar Graph mode
            detailGraphBtn_Click(null, EventArgs.Empty);

            // Set Notify Icon visible
            notifyIconToolStripMenuItem.Checked = true;
            notifyIconToolStripMenuItem_Click(null, EventArgs.Empty);

            this.Text = this.notifyIcon.Text = ProductNameAndVersion();
            this.notifyMenuTitle.ToolTipText = ProductNameAndVersion();

            this.mapCombo.Text = this.mapCombo.Items[0].ToString();

            // this.mapZoom = this.mapBox.Size;
            this.mapZoom = new Size(this.mapImage.Width, this.mapImage.Width * mapBox.Height / mapBox.Width);
            this.mapBox.BackgroundImage = new Bitmap(mapBox.Width, mapBox.Height);
            this.mapBox.Image = new Bitmap(mapBox.Width, mapBox.Height);
            this.mapBoxBeforeZoom = this.mapBox.Size;
            Size mapBoxOffset = new Size(mapBox.Parent.Width - mapBox.Width, mapBox.Parent.Height - mapBox.Height);

            this.mapMinXBox.Text = mapMinXdeg.ToString();
            this.mapMinYBox.Text = mapMinYdeg.ToString();
            this.mapMaxXBox.Text = mapMaxXdeg.ToString();
            this.mapMaxYBox.Text = mapMaxYdeg.ToString();

            PanImage(mapImage, mapOrigin);

            Size imgSize = new Size(18, 18);
            markColorBtn.Image = new Bitmap(imgSize.Width, imgSize.Height);
            markColorBtn_Click(null, EventArgs.Empty);

            InitData();
        }

        #region ==== Data Members

        PrintListView printListView = new PrintListView();

        [Serializable]
        internal class Station
        {
            public Station(ListViewItem item)
            {
                country = item.SubItems[0].Text;
                state = item.SubItems[1].Text;
                place = item.SubItems[2].Text;
                icao = item.SubItems[3].Text;
                latitude.Text = item.SubItems[4].Text;
                longitude.Text = item.SubItems[5].Text;
                enable = item.Checked;
            }

            public Station()
            {
                country = state = place = icao = string.Empty;
                enable = false;
            }

            public void FillItem(ListViewItem item)
            {
                while (item.SubItems.Count < 6)
                    item.SubItems.Add(string.Empty);

                item.SubItems[0].Text = country;
                item.SubItems[1].Text = state;
                item.SubItems[2].Text = place;
                item.SubItems[3].Text = icao;
                item.SubItems[4].Text = latitude.Text;
                item.SubItems[5].Text = longitude.Text;
                item.Checked = enable;
                item.Tag = this;
            }

            public string country;
            public string state;
            public string place;
            public string icao;
            public Angle latitude = new Angle("F2");
            public Angle longitude = new Angle("F2");
            public bool enable;
        }

        [Serializable]
        internal class Presentation
        {
            public Presentation() 
            {
                Init(Pstyle.english, "F1");
            }

            // public TimeZoneInfo tzinfo;
            // public TimeZone tz;
            public WxString country;
            public WxString state;
            public WxString place;
            public WxString icao;
            public WxTime time;
            public Temperature temperature;
            public Temperature dewpoint;
            public Wind wind;
            public Pressure pressure;
            public Visibility visibility;
            public Precipitation precipitation;
            public Clouds clouds;
            public Weather weather;
            public WxTime pulled;
            public WxTimeSpan age;
            public Angle latlon;

            public enum Pstyle { english, metric };

            public void Init(Pstyle pstyle, string fFmt)
            {
                country = new WxString();
                state = new WxString();
                place = new WxString();
                icao = new WxString();
                time = new WxTime();

                temperature = new Temperature();
                dewpoint = new Temperature();
                wind = new Wind();
                pressure = new Pressure();
                visibility = new Visibility();
                precipitation = new Precipitation();

                clouds = new Clouds();
                weather = new Weather();

                pulled = new WxTime();
                age = new WxTimeSpan();
                latlon = new Angle("F3");

                switch (pstyle)
                {
                    case Pstyle.english:
                        temperature.Unit = (int)Temperature.Units.Farhrenheit;
                        temperature.Format = fFmt;
                        dewpoint = temperature;
                        wind.Speed.Unit = (int)Speed.Units.MilesPerHour;
                        wind.Gust = wind.Speed;
                        visibility.Unit = (int)LDistance.Units.Mile;
                        visibility.Format = fFmt;
                        precipitation.Unit = (int)Precipitation.Units.Inches;
                        precipitation.Format = fFmt;

                        // SetStyle(shortText, longText, icon, agif)
                        // clouds
                        // weather
                        break;
                    case Pstyle.metric:
                        temperature.Unit = (int)Temperature.Units.Celcius;
                        temperature.Format = fFmt;
                        dewpoint = temperature;
                        wind.Speed.Unit = (int)Speed.Units.KmPerHour;
                        wind.Gust = wind.Speed;
                        visibility.Unit = (int)LDistance.Units.Kilometer;
                        visibility.Format = fFmt;
                        precipitation.Unit = (int)Precipitation.Units.Centimeters;
                        precipitation.Format = fFmt;

                        // SetStyle(shortText, longText, icon, agif)
                        // clouds
                        // weather
                        break;
                }
            }
        }

        [Serializable]
        internal class ColumnOrder
        {
            public ColumnOrder()
            {
            }

            [Serializable]
            public class ColOrdInfo
            {
                public ColOrdInfo()
                {
                    index = dspIdx = width = 0;
                    text = string.Empty;
                }
                public int index;
                public int dspIdx;
                public int width;
                public string text;
            }

            public class DspCompare : IComparer<ColOrdInfo>
            {
                public int Compare(ColOrdInfo x, ColOrdInfo y)
                {
                    return x.dspIdx - y.dspIdx;
                }
            }
            public class IdxCompare : IComparer<ColOrdInfo>
            {
                public int Compare(ColOrdInfo x, ColOrdInfo y)
                {
                    return x.index - y.index;
                }
            }

            public List<ColOrdInfo> m_colOrdInfoList = new List<ColOrdInfo>();

            public void SetViewOrder(ListView listView)
            {
                int colCnt = listView.Columns.Count;
                for (int colIdx = 0; colIdx < colCnt; colIdx++)
                {
                    if (colIdx < m_colOrdInfoList.Count)
                    {
                        ColOrdInfo colOrdInfo = m_colOrdInfoList[colIdx];
                        listView.Columns[colIdx].DisplayIndex = colOrdInfo.dspIdx;
                        listView.Columns[colIdx].Width = colOrdInfo.width;
                    }
                }
            }

            public void GetViewOrder(ListView listView)
            {
                m_colOrdInfoList.Clear();

                foreach (ColumnHeader colHdr in listView.Columns)
                {
                    ColOrdInfo colOrdInfo =  new ColOrdInfo{
                        index = colHdr.Index, 
                        dspIdx = colHdr.DisplayIndex,
                        width = colHdr.Width,
                        text = colHdr.Text};
                    m_colOrdInfoList.Add(colOrdInfo);
                }
            }
        }

        [Serializable]
        internal class UrlData
        {
            public string dtype;
            public string url;
            public bool enabled;
        }

        [Serializable]
        internal class Sound
        {
            public enum Action { WebFetch, WebError, WebResponse, ClockHour, MetarNew } ;
            public bool Enabled;
            public Action action;
            public string wavPath;
        }

        [Serializable]
        internal class AgeColor
        {
            public TimeSpan age;
            public Color bgColor;
            public Color fgColor;
        }

        [Serializable]
        internal class MapSpecs
        {
            public string name;
            public string path;
            public string bounds;   // minX minY maxX maxY
        }

        [Serializable]
        internal class Config
        {
            public Config()
            {
                webFetchInterval = new TimeSpan(0, 5, 0);
                sounds = new List<Sound>();
                ageColors = new SortedList<TimeSpan, AgeColor>();
            }

            public TimeSpan webFetchInterval;
            public List<Sound> sounds;
            public SortedList<TimeSpan, AgeColor> ageColors;
            public List<MapSpecs> mapSpecs;
        }

        [Serializable]
        internal class Bookmark
        {
            public string name;
            public Color color;
            public Keys keycode;
            public string description;

            public List<Station> stations;
            public Presentation presentation;
        }

        [Serializable]
        public class ColumnObj
        {
            public ColumnObj(int colIdx, WxBase wxb)
            {
                m_colIdx = colIdx;
                m_wxbase = wxb;
            }


            public int m_colIdx;
            public WxBase m_wxbase;
        }

        [Serializable]
        internal class Settings
        {
            public Settings()
            {
                stations = new List<Station>();
                presentation = new Presentation();
                columnOrder = new ColumnOrder();
                urlList = new List<UrlData>();
                waitSettings = 60;
            }

            public List<Station> stations;
            public Presentation presentation;
            public ColumnOrder columnOrder;
            public List<UrlData> urlList;
            public int waitSettings;
            public Config config;
            public List<Bookmark> bookmarks;
            public ColumnObj graphColumn;

            // [NonSerialized]
        };

        // TODO - move into settings.
        int maxLogItems = 100;
        int maxWebItems = 100;

        Settings settings = new Settings();
 
        #region ==== Thread Data
        const int ThreadIdIdle = 0;
        const int ThreadIdDone = -1;
        const int ThreadIdAbort = -2;
        bool threadLoadMetarsAbort = false;
        List<Thread> threads = new List<Thread>();
        #endregion 

        #region ==== Message/Error Logging Data
        Color MsgColorErr = Color.LightPink;
        Color MsgColorBeg = Color.LightYellow;
        Color MsgColorEnd = Color.LightGreen;

        class MsgObj
        {
            public MsgObj(string n, string m, Color c)
            {
                dateTime = DateTime.Now;
                name = n;
                message = m;
                color = c;
            }

            public DateTime dateTime;
            public string name;
            public string message;
            public Color color;
        }

        Queue<MsgObj> msgQueue = new Queue<MsgObj>();
        #endregion

        #region ==== Url data
        string urlRaw = "ftp://tgftp.nws.noaa.gov/data/observations/metar/stations/{0}.TXT";
        string urlTxt1 = "ftp://tgftp.nws.noaa.gov/data/observations/metar/decoded/{0}.TXT";
        string urlArc1 = "ftp://tgftp.nws.noaa.gov/data/observations/metar/cycles/(HH)Z.TXT";
        // string urlArc1 = "ftp://tgftp.nws.noaa.gov/data/observations/metar/cycles/06Z.TXT";
        string urlHtml1 = "http://weather.noaa.gov/weather/current/{0}.html";

        class Url2
        {
            public Url2(string u, string t, Station s)
            {
                url = u;
                typ = t;
                station = s;
            }

            public string url;
            public string typ;
            public DateTime dateTime;
            public Station station;
            public int threadId;
        }
        List<Url2> urlList = new List<Url2>();
        #endregion

        #region ==== web Data Response Data
        class WebData
        {
            public WebData(Url2 _url2, string _str)
            {
                url2 = _url2;
                str = _str;
            }

            public Url2 url2;
            public string str;
        }

        List<WebData> webDataList = new List<WebData>();
        #endregion

     
        bool columnReordered = false;
        bool testMode = true;   // true while no data has been fetched, allow graphs to draw test data.
        List<Metar> newMetarList = new List<Metar>();
        List<Metar> oldMetarList = new List<Metar>();
        Dictionary<string, Station> stationList = new Dictionary<string, Station>();
        DateTime lastUpdate = DateTime.Now;

        const string decodeStr = "decode";
        const string xmlStr = "xml";
        const string txtStr = "text";
        const string htmlStr = "html";
        const string arcStr = "archive"; // collection of reports from the past

        private System.Media.SoundPlayer sPlayer = new System.Media.SoundPlayer();

        #endregion

        #region ==== Basic UI buttons
        void InitData()
        {

            HttpWebRequest.DefaultCachePolicy = new System.Net.Cache.RequestCachePolicy(System.Net.Cache.RequestCacheLevel.CacheIfAvailable);
          
            // "http://www.nzone.com/object/nzone_hdcontent_home.html"
            int defConnections = System.Net.ServicePointManager.DefaultConnectionLimit;
            System.Net.ServicePointManager.DefaultConnectionLimit = 20;

            RestoreSettings();

            foreach (UrlData urlData in settings.urlList)
            {
                ListViewItem urlItem;
                urlItem = this.urlView.Items.Add(urlData.dtype);
                urlItem.SubItems.Add(urlData.url);
                urlItem.Checked = urlData.enabled;
            }

            foreach (ListViewItem urlItem in urlView.Items)
            {
                string urlStr = urlItem.SubItems[1].Text;
                comboBro.Items.Add(urlStr.Substring(0, urlStr.LastIndexOf('/')+1));
            }

            comboBro.Items.Add("http://www.wunderground.com/wundermap/");
            comboBro.Items.Add("http://www.easymetar.com/");
            comboBro.Items.Add("http://aviationweather.gov/adds/metars/");
            comboBro.Items.Add("http://www.rap.ucar.edu/weather/surface/");
            comboBro.Items.Add("http://www.metarsdownloader.com/");
            comboBro.Items.Add("http://www.orbifly.com/member/metmap.php?region_choose=UKI&mode=metar&lang=ENG");


            FillUnitView();
            UpdateUnitView();

            threadWaitProgress.Visible = false;
            threadWaitProgress.Maximum = settings.waitSettings;

            Assembly a = Assembly.GetExecutingAssembly();
            Stream zipStream = a.GetManifestResourceStream("WebMetar_NS.mststatn2010.dat.gz");
            GZipStream decompressStream = new GZipStream(zipStream, CompressionMode.Decompress, false);
            TextReader txtReader = new StreamReader(decompressStream);
            LoadMasterStation(txtReader);
        }

        public static string ProductNameAndVersion()
        {
            string appName = Application.ProductName;
            string appVern = Application.ProductVersion;
            return appName + " v" + appVern.Substring(0, 3); //  Get part of versoin string "n.n"
        }

        string PrintTitle()
        {
            return string.Format("{0}     "
                    , this.Text
                // , DateTime.Now.ToString("G")
                // , Dns.GetHostName()
                // , System.Environment.UserName
                // , System.Environment.OSVersion.ToString()
                    );
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            try
            {
                SaveSettings();

                threadLoadMetarsAbort = notifyIcon.Visible;
                foreach (Thread threadLoadMetars in threads)
                {
                    if (threadLoadMetars.IsAlive)
                    {
                        threadLoadMetars.Abort();
                        threadLoadMetars.Join();
                    }
                }
            }
            catch (Exception ex) 
            {
                MessageBox.Show("close error " + ex.Message);
            }

            this.Visible = false;
            e.Cancel = notifyIcon.Visible;
            base.OnClosing(e);
        }

        private void SortGroup(ListViewGroup lvGroup)
        {
            ListView lvSort = new ListView();

            // Cannot store lvItem in two lists, so move it.
            for (int idx = lvGroup.Items.Count - 1; idx >= 0 ; idx--)
            {
                ListViewItem item = lvGroup.Items[idx];
                lvGroup.ListView.Items.RemoveAt(item.Index);
                lvSort.Items.Add(item);
            }

            lvSort.ListViewItemSorter = lvGroup.ListView.ListViewItemSorter;
            lvSort.Sort();

            // Cannot store lvItem in two lists, move back to original list.
            int cnt = lvSort.Items.Count;
            for (int idx = 0; idx < cnt; idx++)
            {
                ListViewItem item = lvSort.Items[0];
                lvSort.Items.Remove(item);
                lvGroup.ListView.Items.Add(item).Group = lvGroup;
            }
        }

        /// <summary>
        /// Column header click fires sort.
        /// Supports two tier sorting.
        /// </summary>
        private void ColumnSort_Click(object sender, ColumnClickEventArgs e)
        {
            bool isCongtrol = System.Windows.Forms.Control.ModifierKeys == Keys.Control;

            ListView listView = sender as ListView;
            ListViewColumnSorter sorter = listView.ListViewItemSorter as ListViewColumnSorter;
            if (sorter == null)
                return;

            if (listView.Groups.Count != 0)
            {
                // MessageBox.Show("Sorting does not work when in 'Group' mode");
                foreach (ListViewGroup lvGroup in listView.Groups)
                    SortGroup(lvGroup);
            }

            // Determine if clicked column is already the column that is being sorted.
            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == sorter.SortColumn1)
            {
                // Reverse the current sort direction for this column.
                if (sorter.Order1 == SortOrder.Ascending)
                    sorter.Order1 = SortOrder.Descending;
                else
                    sorter.Order1 = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                sorter.SortColumn2 = sorter.SortColumn1;
                sorter.Order2 = sorter.Order1;
                sorter.SortColumn1 = e.Column;
                sorter.Order1 = SortOrder.Ascending;
            }

            // Clear old arrows and set new arrow
            foreach (ColumnHeader colHdr in listView.Columns)
                colHdr.ImageIndex = -1;

            if (sorter.SortColumn1 != sorter.SortColumn2)
                listView.Columns[sorter.SortColumn2].ImageIndex = (sorter.Order2 == SortOrder.Ascending) ? 2 : 3;

            listView.Columns[sorter.SortColumn1].ImageIndex = (sorter.Order1 == SortOrder.Ascending) ? 0 : 1;

            sorter.SortCheckedState = (isCongtrol && sorter.SortColumn1 == 0);

            sorter.SortType =
                (listView.Columns[sorter.SortColumn1].TextAlign == HorizontalAlignment.Right) ?
                ListViewColumnSorter.SortDataType.eNumeric :
                ListViewColumnSorter.SortDataType.eAlpha;

            // Perform the sort with these new sort options.
            if (listView != null)
                listView.Sort();
        }

        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void exitMenu_Click(object sender, EventArgs e)
        {
            notifyIcon.Visible = false;
            this.Close();
        }

        private void aboutMenu_Click(object sender, EventArgs e)
        {
            About about = new About();
            about.Show();
        }

        private void printSetupMenu_Click(object sender, EventArgs e)
        {
            printListView.PageSetup();
        }

        private void runningMenuCk_Click(object sender, EventArgs e)
        {
            if (runningMenuCk.Checked)
            {
                runningMenuCk.Text = "Pause";
                pauseToolStripMenuItem.Text = "Pause";
                this.pauseToolStripMenuItem.ToolTipText =
                        string.Format("Pause [{0} of {1}]",
                        settings.waitSettings - threadWaitProgress.Value,
                        settings.waitSettings);
                threadLoadMetarsAbort = false;
                tableToolStrip.BackColor = Color.LightGreen;
                threadWaitProgress.Visible = true;
            }
            else
            {
                runningMenuCk.Text = "Running";
                pauseToolStripMenuItem.Text = "Run";
                this.pauseToolStripMenuItem.ToolTipText = "Enable auto load";
                tableToolStrip.BackColor = Color.LightBlue;
                threadWaitProgress.Visible = false;
            }
        }
        #endregion

        #region ==== Settings
        const string SettingsFile = @"\webmetar.dat";

        public void SaveSettings()
        {
            // First populate Settings with our stuff
            settings.stations = new List<Station>(stationListView.Items.Count);
            foreach (ListViewItem staItem in stationListView.Items)
            {
                settings.stations.Add(new Station(staItem));
            }

            settings.columnOrder.GetViewOrder(this.tableDataView);

            settings.bookmarks.Clear();
            foreach (ListViewItem bmItem in bookmarkView.Items)
            {
                Bookmark bookmark = (Bookmark)bmItem.Tag;
                settings.bookmarks.Add(bookmark);
            }

            string ourStuff = System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
            ourStuff += SettingsFile;

            try
            {
                Stream stream = File.OpenWrite(ourStuff);
                BinaryFormatter bFormatter = new BinaryFormatter();
                bFormatter.Serialize(stream, settings);
                stream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("failed to save settings webmetar.dat " + ex.Message);
            }
        }

        public void UpdateStationView()
        {
            stationListView.Items.Clear();
            foreach (Station station in settings.stations)
            {
                ListViewItem item = stationListView.Items.Add(string.Empty);
                station.FillItem(item);
            }

            UpdateStationStatus();
        }

        public void RestoreSettings()
        {
            string ourStuff = System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
            ourStuff += SettingsFile;

            try
            {
                Stream stream = File.OpenRead(ourStuff);
                try
                {
                    BinaryFormatter bFormatter = new BinaryFormatter();
                    settings = (Settings)bFormatter.Deserialize(stream);

                    // Restore stuff
                    UpdateStationView();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Sorry saved settings reset");
                }
                stream.Close();
            }
            catch (Exception ex)
            {
                // MessageBox.Show("Failed to open webmetar.dat " + ex.Message);
            }

            // Init any missing parts
            if (settings.presentation == null)
                settings.presentation = new Presentation();
            if (settings.columnOrder == null)
                settings.columnOrder = new ColumnOrder();
            if (settings.columnOrder.m_colOrdInfoList.Count == 0)
                settings.columnOrder.GetViewOrder(this.tableDataView);
            if (settings.bookmarks == null)
                settings.bookmarks = new List<Bookmark>();
            if (settings.graphColumn == null)
                settings.graphColumn = new ColumnObj(5, settings.presentation.temperature);

            // Restore data table view order
            settings.columnOrder.SetViewOrder(this.tableDataView);

            if (settings.urlList == null || settings.urlList.Count == 0)
            {
                settings.urlList = new List<UrlData>();
                UrlData url1 = new UrlData();
                url1.dtype = decodeStr;
                url1.url = urlRaw;
                url1.enabled = true;
                settings.urlList.Add(url1);

                UrlData url2 = new UrlData();
                url2.dtype = txtStr;
                url2.url = urlTxt1;
                url2.enabled = false;
                settings.urlList.Add(url2);

                UrlData url3 = new UrlData();
                url3.dtype = htmlStr;
                url3.url = urlHtml1;
                url3.enabled = false;
                settings.urlList.Add(url3);

                UrlData url4 = new UrlData();
                url4.dtype = arcStr;
                url4.url = urlArc1;
                url4.enabled = false;
                settings.urlList.Add(url4);
            }

            if (settings.config == null || settings.config.sounds == null || settings.config.ageColors == null || settings.config.mapSpecs == null)
            {
                settings.config = new Config();

                settings.config.sounds.Clear();
                settings.config.sounds.Add(new Sound() { Enabled = true, action = Sound.Action.ClockHour, wavPath = @"C:\WINDOWS\Media\chimes.wav" });
                settings.config.sounds.Add(new Sound() { Enabled = true, action = Sound.Action.MetarNew, wavPath = @"C:\WINDOWS\Media\notify.wav" });
                settings.config.sounds.Add(new Sound() { Enabled = false, action = Sound.Action.WebFetch, wavPath = @"C:\WINDOWS\Media\ding.wav" });
                settings.config.sounds.Add(new Sound() { Enabled = true, action = Sound.Action.WebResponse, wavPath = @"C:\WINDOWS\Media\ding.wav" });
                settings.config.sounds.Add(new Sound() { Enabled = true, action = Sound.Action.WebError, wavPath = @"C:\WINDOWS\Media\Windows XP Error.wav" });

                settings.config.ageColors.Clear();
                settings.config.ageColors.Add(new TimeSpan(0, 5, 0), new AgeColor() { bgColor = Color.LightGreen, fgColor = Color.Black });
                settings.config.ageColors.Add(new TimeSpan(0, 15, 0), new AgeColor() { bgColor = Color.Transparent, fgColor = Color.Green });
                settings.config.ageColors.Add(new TimeSpan(0, 30, 0), new AgeColor() { bgColor = Color.Transparent, fgColor = Color.Black });
                settings.config.ageColors.Add(new TimeSpan(2, 0, 0), new AgeColor() { bgColor = Color.Transparent, fgColor = Color.Black });
                settings.config.ageColors.Add(new TimeSpan(12, 0, 0), new AgeColor() { bgColor = Color.LightPink, fgColor = Color.Black });
                settings.config.ageColors.Add(TimeSpan.MaxValue, new AgeColor() { bgColor = Color.Red, fgColor = Color.White });
            }
            // TODO - move back into if-statement
            {
                settings.config.mapSpecs = new List<MapSpecs>();
                settings.config.mapSpecs.Add(new MapSpecs()
                {
                    name = unitedStates48Str,
                    path = internalMapStr, 
                    bounds = "-125.0 25.0 -66.8 49.7" });
                settings.config.mapSpecs.Add(new MapSpecs() { name = worldStr, path = internalMapStr, bounds = "-180 -90 180 84" });
            }

            foreach (Sound sound in settings.config.sounds)
            {
                string action = Enum.GetName(typeof(Sound.Action), sound.action);
                ListViewItem wavItem = cfgSoundView.Items.Add("");
                wavItem.SubItems.Add(action);
                wavItem.SubItems.Add(sound.wavPath);
                wavItem.Checked = sound.Enabled;
            }

            foreach (Bookmark bookmark in settings.bookmarks)
            {
                ListViewItem bmItem = this.bookmarkView.Items.Add("");
                bmItem.UseItemStyleForSubItems = false;
                bmItem.SubItems.Add(bookmark.name);
                bmItem.SubItems.Add("").BackColor = bookmark.color;
                bmItem.SubItems.Add("no Key");
                bmItem.SubItems.Add(bookmark.description);
                bmItem.Tag = bookmark;
            }

            this.mapCombo.Items.Clear();
            this.mapListView.Items.Clear();
            foreach (MapSpecs mapSpecs in settings.config.mapSpecs)
            {
                this.mapCombo.Items.Add(mapSpecs.name);
                ListViewItem mapItem = this.mapListView.Items.Add(mapSpecs.name);
                mapItem.SubItems.Add(mapSpecs.path);
                mapItem.SubItems.Add(mapSpecs.bounds);
            }

            UpdateAgeView();
         
            cfgWebFetchSettings.Value = settings.waitSettings;
        }

        private void UpdateAgeView()
        {
            ageColorView.Items.Clear();

            foreach (TimeSpan age in settings.config.ageColors.Keys)
            {
                AgeColor ageColor = settings.config.ageColors[age];

                ListViewItem ageItem = ageColorView.Items.Add(TimeSpanToString(age));
                ageItem.UseItemStyleForSubItems = false;
                ageItem.SubItems.Add("");
                ageItem.SubItems[1].BackColor = ageColor.fgColor;
                ageItem.SubItems.Add("");
                ageItem.SubItems[2].BackColor = ageColor.bgColor;
                ageItem.Tag = age;
            }
        }

        string TimeSpanToString(TimeSpan age)
        {
            if (age == TimeSpan.MinValue)
                return "hh:mm";
            if (age == TimeSpan.MaxValue)
                return "Max";

            return string.Format("{0:00}:{1:00}:{2:00}", age.Hours, age.Minutes, age.Seconds);
        }

        TimeSpan ParseTimeSpan(string text)
        {
            if (string.Compare(text, "max", true) == 0)
            {
                return TimeSpan.MaxValue;
            }

            try
            {
                string[] hhmmss = text.Trim().Split(':');
                int hour = int.Parse(hhmmss[0]);
                int min = (hhmmss.Length > 1) ? int.Parse(hhmmss[1]) : 0;
                int sec = (hhmmss.Length > 2) ? int.Parse(hhmmss[2]) : 0;
                return new TimeSpan(hour, min, sec);
            }
            catch
            {
                return TimeSpan.MinValue;
            }
        }
        #endregion

        #region ==== Timer Controls
        int milliSeconds = 0;
       
        int lastHour = DateTime.Now.Hour;
        static string lastWave = string.Empty;
        private System.Media.SoundPlayer wavPlayer = new System.Media.SoundPlayer();


        private void PlaySound(Sound.Action action)
        {
            foreach (Sound sound in settings.config.sounds)
            {
                if (sound.action == action && sound.Enabled)
                {
                    try
                    {
                        if (sound.wavPath.Length > 0 && File.Exists(sound.wavPath))
                        {
                            if (lastWave != sound.wavPath)
                            {
                                lastWave = sound.wavPath;
                                wavPlayer.Stream = new FileStream(sound.wavPath, FileMode.Open, FileAccess.Read, FileShare.Read);
                            }

                            wavPlayer.Play();
                            return;
                        }
                    }
                    catch { }
                    break;
                }
            }
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            if (DateTime.Now.Hour != lastHour)
            {
                lastHour = DateTime.Now.Hour;
                PlaySound(Sound.Action.ClockHour);
            }

            if (columnReordered)
                ColumnReordered();

            lock (msgQueue)
            {
                while (msgQueue.Count != 0)
                {
                    MsgObj logMsg = msgQueue.Dequeue();
                    ListViewItem item = null;
                    item = logView.Items.Add(logMsg.dateTime.ToString("HH:mm:ss.fff"));
                    item.SubItems.Add(logMsg.name);
                    item.SubItems.Add(logMsg.message);
                    item.BackColor = logMsg.color;
                    item.EnsureVisible();
                }

                while (logView.Items.Count > maxLogItems)
                {
                    logView.Items.RemoveAt(0);
                }
                logStatus.Text = string.Format("#Lines:{0:#,##0}", logView.Items.Count);
            }

            lock (webDataList)
            {
                foreach (WebData webData in webDataList)
                {
                    ListViewItem webItem = this.webResponseView.Items.Add(webData.url2.dateTime.ToString("HH:mm:ss.fff"));
                    webItem.SubItems.Add(webData.url2.url);
                    webItem.SubItems.Add(webData.str);
                }

                while (webResponseView.Items.Count > maxWebItems)
                {
                    webResponseView.Items.RemoveAt(0);
                }
                webStatus.Text = string.Format("#Lines:{0:#,##0}", webResponseView.Items.Count);

                webDataList.Clear();
            }

            lock (newMetarList)
            {
                // Age list.
                foreach (ListViewItem metarItem in tableDataView.Items)
                {
                    Metar metar = (Metar)metarItem.Tag;

#if true
                    // Valid time age
                    DateTime dt = metar.reportTimeUTC;
                    TimeSpan itemAge = DateTime.UtcNow - dt;
#else
                    // Pull age
                    DateTime dt = (metarItem.SubItems.Count > cold_pulled.Index) ?
                        (DateTime)metarItem.SubItems[cold_pulled.Index].Tag :
                        DateTime.Now;
                    TimeSpan itemAge = dt - DateTime.Now;
#endif
                    settings.presentation.age.SetValue(itemAge.Ticks, 0);
                    metarItem.SubItems[cold_age.Index].Text = settings.presentation.age.Text;

                    AgeColor ageColor = null;
                    foreach (TimeSpan age in settings.config.ageColors.Keys)
                    {
                        if (itemAge <= age)
                        {
                            ageColor = settings.config.ageColors[age];
                            break;
                        }
                    }

                    if (ageColor != null)
                    {
                        // metarItem.BackColor = Towards(tableDataView.BackColor, metarItem.BackColor, 1);
                        metarItem.BackColor = 
                            (ageColor.bgColor == Color.Transparent) ? tableDataView.BackColor : ageColor.bgColor;
                        metarItem.ForeColor =
                            (ageColor.fgColor == Color.Transparent) ? tableDataView.ForeColor : ageColor.fgColor;
                    }
                }

                int newMetarCnt = 0;
                foreach (Metar metar in newMetarList)
                {
                    if (false == Find(metar, oldMetarList))
                    {
                        testMode = false;
                        oldMetarList.Add(metar);
                        newMetarCnt++;
                        UpdateMetarView(metar);
                    }
                }

                newMetarList.Clear();

                if (newMetarCnt != 0)
                {
                    PlaySound(Sound.Action.MetarNew);
                }
            }

            milliSeconds += timer.Interval;
            if (threadWaitProgress.Visible && milliSeconds / 1000 < threadWaitProgress.Maximum)
            {
                threadWaitProgress.Value = milliSeconds / 1000;
                this.threadWaitProgress.ToolTipText = 
                        string.Format("{0} seconds until next background web fetch.", 
                        settings.waitSettings - threadWaitProgress.Value);
                this.pauseToolStripMenuItem.ToolTipText = 
                        string.Format("Pause [{0} of {1}]",
                        settings.waitSettings - threadWaitProgress.Value,
                        settings.waitSettings);
            }
            if (milliSeconds > settings.waitSettings * 1000)          
            {
                milliSeconds = 0;
                if (runningMenuCk.Checked)
                {
                    GetWebResponses();
                }
            }

            localTmTblLbl.Text = DateTime.Now.ToString("HH:mm:ss") 
                + char.ToUpper(TimeZone.CurrentTimeZone.StandardName[0]);
            gmtTmTblLbl.Text = DateTime.UtcNow.ToString("HH:mm:ssZ");
        }
        #endregion

        #region ==== Metar
        SortedList<string, ListViewGroup> tableGroup = new SortedList<string, ListViewGroup>();

        private void UpdateMetarView(Metar metar)
        {
            ListViewItem metarItem = tableDataView.Items.Add("");
            // metarItem.BackColor = Color.LightGreen;

            Station station = FindStation(metar.icao);
            if (station == null)
                return;     // Should never get here

            ListViewGroup grp = null;
            if (viewTblGrpIcaoMenu.Checked)
            {
                if (tableGroup.TryGetValue(metar.icao, out grp) == false)
                {
                    grp = tableDataView.Groups.Add(metar.icao, metar.icao);
                    tableGroup.Add(metar.icao, grp);
                }
            }
            else if (viewTblGrpHourMenu.Checked)
            {
                string timeStr = string.Format("{0} Hour:{1}",
                    metar.reportTimeUTC.ToString("yyyy-MMM-dd"),
                    metar.reportTimeUTC.ToString("HH (GMT)"));
                if (tableGroup.TryGetValue(timeStr, out grp) == false)
                {
                    grp = tableDataView.Groups.Add(timeStr, timeStr);
                    tableGroup.Add(timeStr, grp);
                }
            }

            metarItem.Group = grp;

            station.FillItem(metarItem);            // 0=cntry, 1=state, 2=place, 3=icao, 4=lat, 5=lon
            metarItem.SubItems.RemoveAt(5);         // remove longitude
            metarItem.SubItems.RemoveAt(4);         // remove latitude

            metarItem.SubItems.Add(metar.reportTimeUTC.ToString(settings.presentation.time.Format));       // 4 = reportTime
            metarItem.SubItems.Add(WxFormat(settings.presentation.temperature, metar.temperature));
            metarItem.SubItems.Add(WxFormat(settings.presentation.dewpoint, metar.dewpoint));
            metarItem.SubItems.Add(WxFormat(settings.presentation.wind.Direction, metar.wind.Direction));
            metarItem.SubItems.Add(WxFormat(settings.presentation.wind.Speed, metar.wind.Speed));
            metarItem.SubItems.Add(WxFormat(settings.presentation.wind.Gust, metar.wind.Gust));
            metarItem.SubItems.Add(WxFormat(settings.presentation.pressure, metar.pressure));
            metarItem.SubItems.Add(WxFormat(settings.presentation.precipitation, metar.precipitation)); // 11 = prec

            metarItem.SubItems.Add((metar.cloudsList.Count != 0) ? WxFormat(settings.presentation.clouds, metar.cloudsList[0]) : string.Empty);
            metarItem.SubItems.Add((metar.visibilityList.Count != 0) ? WxFormat(settings.presentation.visibility, metar.visibilityList[0]) : string.Empty);
            metarItem.SubItems.Add((metar.weatherList.Count != 0) ? WxFormat(settings.presentation.weather, metar.weatherList[0]) : string.Empty);

            DateTime now = DateTime.Now;
            metarItem.SubItems.Add(now.ToString(settings.presentation.pulled.Format)).Tag = now;

            TimeSpan age = DateTime.UtcNow - metar.reportTimeUTC;
            settings.presentation.age.SetValue(age.Ticks, (int)WxTimeSpan.Units.Default);
            metarItem.SubItems.Add(settings.presentation.age.Text).Tag = age;

            metarItem.SubItems.Add(WxFormat(settings.presentation.latlon, station.latitude));
            metarItem.SubItems.Add(WxFormat(settings.presentation.latlon, station.longitude)); 

            metarItem.Tag = metar;
            lastUpdate = DateTime.Now;
            UpdateTableStatus();

            if (tabData.SelectedTab == tabDetail)
            {
                UpdateGraph();
            }
        }

        private void UpdateMetarView()
        {
            // Reset view groups
            tableGroup.Clear();
            tableDataView.Groups.Clear();

            this.tableDataView.Items.Clear();
            foreach (Metar metar in oldMetarList)
            {
                UpdateMetarView(metar);
            }
        }

        private void UpdateBarGraph()
        {
            this.barGraph.Clear();
            int colIdx = settings.graphColumn.m_colIdx;
            ColumnObj columnObj = (ColumnObj)this.tableDataView.Columns[colIdx].Tag;
            string parmName = tableDataView.Columns[colIdx].Text;
            string unitText = columnObj.m_wxbase.UnitText;
            
            this.barGraph.BarSettings.HTitle = parmName + " " + unitText;
            Size sz = this.barGraph.BarSettings.BarSize;
            this.barGraph.BarSettings.CellSize = sz + sz;

            string[] labels = stationList.Keys.ToArray();
            string[] tip = new string[] { "{2:HH:mm}z {1} " + parmName + " {0:F1}" + unitText };
            List<BarGraph.DataItem> barData = new List<BarGraph.DataItem>();

            foreach (ListViewItem metarItem in this.tableDataView.Items)
            {
                Metar metar = (Metar)metarItem.Tag;
                if (stationList.ContainsKey(metar.icao))
                {
                    double value1 = metar.GetParameter(parmName).GetValue(columnObj.m_wxbase.Unit);
                    // double value1 = metar.temperature.GetValue(this.settings.presentation.temperature.Unit);

                    double[] values = new double[] { value1 };
                    barData.Add(new BarGraph.DataItem(metar.icao, metar.reportTimeUTC, 
                        metarItem.Selected, values));
                }
            }

            this.barGraph.Data(barData, tip);
        }

        private void UpdateTimeGraph()
        {
            this.timeGraph.Clear();
            List<TimeGraph.DataItem> timeData = new List<TimeGraph.DataItem>();
            SortedList<string, bool> stations = new SortedList<string, bool>();

            DateTime baseTime = DateTime.MaxValue;
            foreach (Metar metar in oldMetarList)
            {
                TimeGraph.DataItem di = new TimeGraph.DataItem();
                di.dt = metar.reportTimeUTC;
                di.name = metar.icao;

                timeData.Add(di);
                stations[di.name] = true;
                baseTime = (baseTime < di.dt) ? baseTime : di.dt;
            }

            if (timeData.Count > 0)
            {
                TimeGraph.DataItem di = new TimeGraph.DataItem();
                di.dt = DateTime.UtcNow;
                di.name = "Now";
                timeData.Add(di);
                stations[di.name] = true;

                string[] tip = new string[] { "{1} Time {0} " };
                this.timeGraph.Data(timeData, baseTime, stations.Keys.ToList(), tip);
            }
        }

        private void UpdateGraph()
        {
            if (oldMetarList.Count > 0 || testMode == false)
            {
                UpdateBarGraph();
                UpdateTimeGraph();
            }
        }

        private void UpdateTableStatus()
        {
            this.tableStatus.Text = string.Format("Rows:{0:#,##0}  Last Upd:{1}",
                this.tableDataView.Items.Count,
                this.lastUpdate.ToString("HH:mm:ss"));
            this.Text = string.Format("{0} {1}/{2}", ProductName, this.lastUpdate.ToString("HH:mm"),
                this.tableDataView.Items.Count);
            this.notifyIcon.Text = this.Text;
        }

        private string WxFormat(WxBase presentAs, WxBase data)
        {
            return data.ConvertTo(presentAs.Unit, presentAs.Format).TextandUnit;
        }

        // Example archive file from ftp://tgftp.nws.noaa.gov/data/observations/metar/cycles/05Z.TXT
        //
        //  2010/01/24 04:55
        //  KOLV 240455Z AUTO 14013KT 3SM FEW012 SCT031 BKN046 10/10 A2952 RMK A01 P037
        //
        //  2010/01/24 04:46
        //  KFKA 240446Z AUTO 13007KT 1/2SM FG OVC001 02/01 A2941 RMK AO2 T00180014
        //
        //  2010/01/24 04:55
        //  KOLV 240455Z AUTO 14013KT 3SM FEW012 SCT031 BKN046 10/10 A2952 RMK A01 P037

        private bool ArchiveMetar(Url2 url2, string archiveStr)
        {
            RegCompare dateReg = new RegCompare("([0-9]{4})/([0-9]{2})/([0-9]{2}) +([0-9]{2}):([0-9]{2})[^\n]*");
            RegCompare icaoReg = new RegCompare("([A-Z][A-Z0-9]{3}) ([0-9]{6})Z[^\n]*");
            int pos = 0;
            while (dateReg.IsMatches(archiveStr, pos))
            {
                pos = dateReg.indexes[0] + dateReg.groups[0].Length + 1;
                if (icaoReg.IsMatches(archiveStr, pos))
                {
                    string icao = icaoReg.groups[1];

                    if (stationList.ContainsKey(icao))
                    {
                        DecoderMetar(url2, dateReg.groups[0] + icaoReg.groups[0]);
                    }

                    pos = icaoReg.indexes[0] + icaoReg.groups[0].Length + 1;
                }
            }

            return false;
        }

        //    BOSTON LOGAN INTERNATIONAL, MA, United States (KBOS) 42-22N 071-01W 54M
        //    Jan 13, 2010 - 08:54 PM EST / 2010.01.14 0154 UTC
        //    Wind: from the WNW (300 degrees) at 13 MPH (11 KT):0
        //    Visibility: 10 mile(s):0
        //    Sky conditions: overcast
        //    Temperature: 27.0 F (-2.8 C)
        //    Windchill: 15 F (-9 C):1
        //    Dew Point: 15.1 F (-9.4 C)
        //    Relative Humidity: 60%
        //    Pressure (altimeter): 30.12 in. Hg (1019 hPa)
        //    ob: KBOS 140154Z 30011KT 10SM SCT065 BKN080 OVC110 M03/M09 A3012 RMK AO2 SLP197 T10281094
        //    cycle: 2
        private bool TxtMetar(Url2 url2, string rawMetarStr)
        {
            string cleanMetarStr = rawMetarStr.Replace("\r", " ").Replace("\n", " ").Trim();
            lock (webDataList)
            {
                webDataList.Add(new WebData(url2, rawMetarStr));
            }

            Metar metar = new Metar();
            if (TextToMetar.Decode(rawMetarStr, ref metar))
            {
                metar.notes = rawMetarStr;

                lock (newMetarList)
                {
                    if (stationList.ContainsKey(metar.icao) == false)
                    {
                        stationList.Add(url2.station.icao, url2.station);
                    }

                    if (false == Find(metar, newMetarList))
                    {
                        newMetarList.Add(metar);
                    }
                }
                return true;
            }

            return false;
        }

        private bool DecoderMetar(Url2 url2, string rawMetarStr)
        {
            string cleanMetarStr = rawMetarStr.Replace("\r", " ").Replace("\n", " ").Trim();
            lock (webDataList)
            {
                webDataList.Add(new WebData(url2, rawMetarStr));
            }

            // 2010/01/11 01:54
            // KBOS 110154Z 28009KT 10SM CLR M06/M18 A3013 RMK AO2 SLP201 T10611178
            RegCompare datePat = new RegCompare("^(20[0-9][0-9])/([0-1][0-9])/([0-3][0-9]) ([0-9][0-9]):([0-9][0-9])");
            DateTime refTimeUTC = DateTime.UtcNow;

            if (datePat.IsMatches(cleanMetarStr))
            {
                try
                {
                    int year = int.Parse(datePat.groups[1]);
                    int mon = int.Parse(datePat.groups[2]);
                    int day = int.Parse(datePat.groups[3]);
                    int hour = int.Parse(datePat.groups[4]);
                    int min = int.Parse(datePat.groups[5]);

                    refTimeUTC = new DateTime(year, mon, day, hour, min, 00);
                }
                catch  
                {
                    refTimeUTC = DateTime.Now;
                }
                cleanMetarStr = cleanMetarStr.Substring(datePat.indexes[5] + 2).Trim();
            }
            else
            {
                return false;
            }

            Metar metar = new Metar();
            if (metar.Decode(cleanMetarStr, refTimeUTC))
            {
                lock (newMetarList)
                {
                    if (stationList.ContainsKey(metar.icao) == false)
                    {
                        stationList.Add(url2.station.icao, url2.station);
                    }

                    if (false == Find(metar, newMetarList))
                    {
                        newMetarList.Add(metar);
                    }
                }
                return true;
            }

            return false;
        }

        #endregion

        #region ==== Master Station List
        private void loadMstBtn_Click(object sender, EventArgs e)
        {
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                LoadMasterStation(openFileDialog.FileName);
            }
        }

        //
        // 0   WMO(5)
        // 1   ICAO(4)
        // 2   Place Name
        // 3   State
        // 4   Country
        // 5   Latitude
        // 6   Longitude
        private bool LoadMasterStation(string filename)
        {
            try
            {
                TextReader txtReader = new StreamReader(filename);
                LoadMasterStation(txtReader);
            }
            catch (Exception ex)
            {
            }

            return false;
        }

        private void LoadMasterStation(TextReader txtReader)
        {
            mstatnListView.Items.Clear();
            mstatnListView.ListViewItemSorter = null;
            mstatnListView.BeginUpdate();

            string lineBuffer;

            while ((lineBuffer = txtReader.ReadLine()) != null)
            {
                string[] fields = lineBuffer.Split(';');
                if (fields.Length == 7)
                {
                    ListViewItem item = mstatnListView.Items.Add(fields[4]);   // 0 country
                    item.SubItems.Add(fields[3]);   // 1 state
                    item.SubItems.Add(fields[2]);   // 2 place
                    item.SubItems.Add(fields[1]);   // 3 ICAO 
                    item.SubItems.Add(fields[5]);   // 4 latitude
                    item.SubItems.Add(fields[6]);   // 5 longitude
                }
            }

            mstatnListView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAlpha);
            mstatnListView.EndUpdate();
            mstStatus.Text = string.Format("Stations: {0:#,##0} ", mstatnListView.Items.Count); 
        }

        private void CopyMstToSta()
        {
            if (mstatnListView.SelectedItems.Count > 0)
            {
                foreach (ListViewItem mstItem in mstatnListView.SelectedItems)
                {
                    // ToDo - check for duplicates
                    ListViewItem staItem = stationListView.Items.Add(mstItem.SubItems[0].Text); // country
                    staItem.SubItems.Add(mstItem.SubItems[1].Text);   // state
                    staItem.SubItems.Add(mstItem.SubItems[2].Text);   // place
                    staItem.SubItems.Add(mstItem.SubItems[3].Text);   // icao
                    staItem.SubItems.Add(mstItem.SubItems[4].Text);   // lat
                    staItem.SubItems.Add(mstItem.SubItems[5].Text);   // long
                }
            }
            else
            {
                foreach (ListViewItem mstItem in mstatnListView.CheckedItems)
                {
                    // ToDo - check for duplicates
                    ListViewItem staItem = stationListView.Items.Add(mstItem.SubItems[0].Text); // country
                    staItem.SubItems.Add(mstItem.SubItems[1].Text);   // state
                    staItem.SubItems.Add(mstItem.SubItems[2].Text);   // place
                    staItem.SubItems.Add(mstItem.SubItems[3].Text);   // icao
                    staItem.SubItems.Add(mstItem.SubItems[4].Text);   // lat
                    staItem.SubItems.Add(mstItem.SubItems[5].Text);   // lon
                }
            }

            UpdateStationStatus();
        }
        #endregion

        #region ==== Helper functions
        #region ---- Static functions
        static private bool Find(Metar metar, List<Metar> oldMetars)
        {
            foreach (Metar oldMetar in oldMetars)
            {
                if (metar.metarStr == oldMetar.metarStr)
                {
                    return true;
                }
            }

            return false;
        }

        static private int ToX(int endX, int begX, int rate)
        {
            if (endX >= begX)
                return Math.Min(endX, begX + rate);
            else
                return Math.Max(endX, begX - rate);
        }
        static private Color Towards(Color endColor, Color begColor, int rate)
        {
            return Color.FromArgb(
                ToX(endColor.R, begColor.R, rate),
                ToX(endColor.G, begColor.G, rate),
                ToX(endColor.B, begColor.B, rate)
                );
        }
       
        static string StringFromStream(Stream resStream)
        {
            string tempString = null;
            int count = 0;
            StringBuilder sb = new StringBuilder();
            byte[] buf = new byte[15000];

            do
            {
                // fill the buffer with data
                count = resStream.Read(buf, 0, buf.Length);

                // make sure we read some data
                if (count != 0)
                {
                    // translate from bytes to ASCII text
                    tempString = Encoding.ASCII.GetString(buf, 0, count);

                    // save results
                    sb.Append(tempString);
                }
            }
            while (count > 0); // any more data to read?

            return sb.ToString();
        }
        
        static string GetBaseURL(WebResponse response)
        {
            string fullURI = response.ResponseUri.AbsoluteUri;
            string baseURL = fullURI.Substring(0, fullURI.IndexOf(response.ResponseUri.AbsolutePath));
            return baseURL;
        }
        #endregion

        private Station FindStation(string icao)
        {
            Station station;
            if (stationList.TryGetValue(icao, out station) == false)
            {
                station = null;
            }

            return station;
        }

        #region ---- UI helper functions
        ColorToolTip colorToolTip = new ColorToolTip();

        private void ShowToolTip(object sender)
        {
            ListView listView = (ListView)sender;
            if (listView != null)
            {
                Control c = listView;
                Point p = listView.PointToClient(Control.MousePosition);
                ListViewHitTestInfo hitInfo = listView.HitTest(p);
                if (hitInfo.SubItem != null)
                {
                    int itemHash = hitInfo.Item.GetHashCode();
                    if ((int)colorToolTip.Tag != itemHash)
                    {
                        colorToolTip.Clear();
                        colorToolTip.Tag = itemHash;
                        colorToolTip.Message = hitInfo.SubItem.Text;
                        // colorToolTip.Append(hitInfo.Item.SubItems[0].Text + " = ", Color.Blue, null);
                        // colorToolTip.Append(hitInfo.Item.SubItems[1].Text, Color.Black, null);
                        colorToolTip.Location = c.PointToScreen(new Point(hitInfo.Item.Bounds.Left, hitInfo.Item.Bounds.Bottom + 10));
                    }
                    colorToolTip.Show();
                    colorToolTip.Width = hitInfo.Item.Bounds.Width;
                }
            }
        }

        private void DeleteListView(ListView listView, string prompt, string title)
        {
            if (listView.SelectedItems.Count == 0)
            {
                if (MessageBox.Show(prompt, title, MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    listView.Items.Clear();
                }
            }
            else
            {
                foreach (ListViewItem item in listView.SelectedItems)
                {
                    listView.Items.Remove(item);
                }
            }
        }
 
        #region ==== ListView Editor
        /// <summary>
        /// Listview column field editor.
        /// </summary>
        /// <param name="listView"></param>
        /// <param name="p">ListView coordinates</param>
        private void ListViewEdit(ListView listView, Point p)
        {
            ListViewHitTestInfo hitInfo = listView.HitTest(p);
            ListViewItem.ListViewSubItem subItem = hitInfo.SubItem;
            int subCol = ((subItem == null) ? -1 : hitInfo.Item.SubItems.IndexOf(subItem));

            if (subCol >= 0)
            {
                ListViewEdit(listView, hitInfo.Item, subCol);
            }
        }

        private void ListViewEdit(ListView listView, ListViewItem item, int subCol)
        {
            Control c = listView;
            ListViewItem.ListViewSubItem subItem = item.SubItems[subCol];

            if (listView == this.unitView)
            {
                ColumnObj columnObj = (ColumnObj)item.Tag;

                switch (subCol)
                {
                    case 0: // column Text, not editable
                        break;
                    case 1: // unit 
                        if (columnObj.m_wxbase.UnitArray.Length > 1)
                        {
                            ComboField comboField = new ComboField();
                            comboField.SetChoices(columnObj.m_wxbase.UnitArray);
                            comboField.ComboText = columnObj.m_wxbase.UnitText;
                            comboField.Location = c.PointToScreen(subItem.Bounds.Location);
                            comboField.Show();
                            comboField.Width = listView.Columns[subCol].Width;
                            comboField.Visible = false;

                            if (comboField.ShowDialog() == DialogResult.OK)
                            {
                                int unitIdx = comboField.ComboIndex;
                                columnObj.m_wxbase.Unit = unitIdx;
                                subItem.Text = columnObj.m_wxbase.UnitText;
                                UpdateMetarView();
                            }
                        }
                        break;
                    case 2: // format
                        if (columnObj.m_wxbase.FormatArray.Length > 1)
                        {
                            ComboField comboField = new ComboField();
                            comboField.SetChoices(columnObj.m_wxbase.FormatArray);
                            comboField.ComboText = columnObj.m_wxbase.Format;
                            comboField.Location = c.PointToScreen(subItem.Bounds.Location);
                            comboField.Show();
                            comboField.Width = listView.Columns[subCol].Width;
                            comboField.Visible = false;

                            if (comboField.ShowDialog() == DialogResult.OK)
                            {
                                columnObj.m_wxbase.Format = comboField.ComboText; ;
                                subItem.Text = columnObj.m_wxbase.Format;
                            }
                            UpdateMetarView();
                        }
                        break;
                    case 3: // graph
                        settings.graphColumn = SetUnitGraph(item.SubItems[0].Text);
                        break;
                }
            }
            else if (listView == cfgSoundView)
            {
                switch (subCol)
                {
#if false
// Do not allow user to edit sound action.
                    case 1:
                        {
                            ComboField comboField = new ComboField();
                            comboField.SetChoices(Enum.GetNames(typeof(Sound.Action)));
                            comboField.ComboText = subItem.Text;
                            comboField.Location = c.PointToScreen(subItem.Bounds.Location);
                            comboField.Show();
                            comboField.Width = listView.Columns[subCol].Width;
                            comboField.Visible = false;

                            if (comboField.ShowDialog() == DialogResult.OK)
                            {
                                subItem.Text = comboField.ComboText;
                                int actionIdx = comboField.ComboIndex;
                                settings.config.sounds[item.Index].action = (Sound.Action)actionIdx;
                            }
                        }
                        break;
#endif
                    case 2:
                        {
                            FileBox fileBox = new FileBox();
                            fileBox.FileText = subItem.Text;
                            fileBox.Location = c.PointToScreen(subItem.Bounds.Location);
                            fileBox.Show();
                            fileBox.Width = listView.Columns[subCol].Width;
                            fileBox.Visible = false;

                            if (fileBox.ShowDialog() == DialogResult.OK)
                            {
                                subItem.Text = fileBox.FileText;
                                settings.config.sounds[item.Index].wavPath = fileBox.FileText;
                                PlaySound(settings.config.sounds[item.Index].action);
                            }
                        }
                        break;
                }
            }
            else if (listView == ageColorView)
            {
                TimeSpan age = (TimeSpan)item.Tag;

                switch (subCol)
                {
                    case 0:
                        ListViewFieldEdit(listView, item, subCol);
                        {
                            TimeSpan newAge = ParseTimeSpan(subItem.Text);
                            if (age != newAge && settings.config.ageColors.ContainsKey(newAge) == false)
                            {
                                AgeColor ageColor = new AgeColor();
                                if (settings.config.ageColors.ContainsKey(age))
                                {
                                    ageColor = settings.config.ageColors[age];
                                    settings.config.ageColors.Remove(age);
                                }
                                ageColor.age = newAge;
                                settings.config.ageColors.Add(newAge, ageColor);
                            }
                            UpdateAgeView();
                        }
                        break;
                    case 1:
                        colorDialog.Color = subItem.BackColor;
                        if (colorDialog.ShowDialog() == DialogResult.OK)
                        {
                            subItem.BackColor = colorDialog.Color;
                            settings.config.ageColors[age].fgColor = colorDialog.Color;
                        }
                        break;
                    case 2:
                        colorDialog.Color = subItem.BackColor;
                        if (colorDialog.ShowDialog() == DialogResult.OK)
                        {
                            subItem.BackColor = colorDialog.Color;
                            settings.config.ageColors[age].bgColor = colorDialog.Color;
                        }
                        break;
                }
            }
            else if (listView == bookmarkView)
            {
                Bookmark bookmark = (Bookmark)item.Tag;

                // Bookmark fields:
                //  0 = ono
                //  1 = name
                //  2 = color
                //  3 = key
                //  4 = desc
                switch (subCol)
                {
                    case 0:
                        break;
                    case 1: // name
                        ListViewFieldEdit(listView, item, subCol);
                        bookmark.name = item.SubItems[1].Text;
                        if (item.Checked)
                            bookmarkLabel.Text = bookmark.name;
                        break;
                    case 2: // color
                        colorDialog.Color = subItem.BackColor;
                        if (colorDialog.ShowDialog() == DialogResult.OK)
                        {
                            subItem.BackColor = colorDialog.Color;
                            bookmark.color = colorDialog.Color;
                            if (item.Checked)
                                ActivateBookmark(item);
                        }
                        break;
                    case 3: // key
                        // TODO - get a hot key
                        break;
                    case 4: // desc
                        ListViewFieldEdit(listView, item, subCol);
                        bookmark.description = item.SubItems[4].Text;
                        bookmarkLabel.ToolTipText = bookmark.description;
                        break;
                }
            }
            else
            {
                ListViewFieldEdit(listView, item, subCol);
            }
        }

        private bool ListViewFieldEdit(ListView listView, ListViewItem item, int subCol)
        {
            Control c = listView;
            ListViewItem.ListViewSubItem subItem = item.SubItems[subCol];

            FieldBox fieldBox = new FieldBox();
            fieldBox.FieldText = subItem.Text;
            fieldBox.Location = c.PointToScreen(subItem.Bounds.Location);
            fieldBox.Show();
            fieldBox.Width = item.Bounds.Width;  // reset width
            fieldBox.Visible = false;

            if (fieldBox.ShowDialog() == DialogResult.OK)
            {
                subItem.Text = fieldBox.FieldText;
                return true;
            }
            return false;
        }

        private void ListView_DoubleClick(object sender, EventArgs e)
        {
            Point p = System.Windows.Forms.Control.MousePosition;
            ListView listView = (ListView)sender;
            if (listView != null)
            {
                p = listView.PointToClient(p);
                ListViewEdit(listView, p);
            }
        }

        private void ListViewEdit_MouseDown(object sender, MouseEventArgs e)
        {

        }

        #endregion
        #endregion

        #endregion

        #region ==== Thread URL loader
        /// <summary>
        /// Load and extract text from html page
        /// </summary>
        /// <param name="strURL"></param>
        /// <returns>string text of html page</returns>
        static string ThreadGetHTMLFromURL(string strURL, out string baseURL, MainForm mainForm)
        {
            string retVal = string.Empty;
            baseURL = string.Empty;
            WebResponse response = null;

            try
            {
                // WebRequest request = WebRequest.Create(strURL);
                if (strURL.StartsWith("ftp", StringComparison.OrdinalIgnoreCase))
                {
                    FtpWebRequest request = (FtpWebRequest)FtpWebRequest.Create(strURL);
                    request.Timeout = 5000;
                    request.UseBinary = false;
                    response = request.GetResponse();
                }
                else
                {
                    HttpWebRequest request = (HttpWebRequest)HttpWebRequest.Create(strURL);
                    request.Timeout = 5000;
                    response = request.GetResponse();
                }

                Stream resStream = response.GetResponseStream();
                retVal = StringFromStream(resStream);
                baseURL = GetBaseURL(response);
            }
            catch (Exception ex)
            {
                lock (mainForm.msgQueue)
                {
                    mainForm.PlaySound(Sound.Action.WebError);
                    mainForm.msgQueue.Enqueue(new MsgObj("web", ex.Message + " " + strURL, mainForm.MsgColorErr));
                }
            }

            if (response != null)
                response.Close();

            return retVal;
        }

        private void StartLoadThread(int startUrl, int numUrl)
        {
            ThreadStart ts = delegate() { ThreadLoadUrl(startUrl, numUrl, this); };
            Thread threadLoadMetars = new Thread(ts);

            for (int idx = threads.Count-1; idx >= 0; idx--)
            {
                if (false == threads[idx].IsAlive)
                {
                    threads.RemoveAt(idx);
                }
            }

            PlaySound(Sound.Action.WebFetch);
            threads.Add(threadLoadMetars);
            threadLoadMetars.Start();
            statusBar.Text = "Threads=" + threads.Count.ToString();
        }

        private void ThreadLoadUrl(int urlIdx, int numUrl, MainForm mainForm)
        {
            bool more = true;
            while (more && false == threadLoadMetarsAbort)
            {
                Url2  url2;
                lock (urlList)
                {
                    more = true;
                    if (urlIdx >= urlList.Count)
                        break;
                    url2 = urlList[urlIdx];
                }

                lock (msgQueue)
                {
                    msgQueue.Enqueue(new MsgObj("Load", url2.url, MsgColorBeg));
                }

                string baseURL;
                string htmlPage = ThreadGetHTMLFromURL(url2.url, out baseURL, mainForm);

                lock (msgQueue)
                {
                    msgQueue.Enqueue(new MsgObj("Load"+htmlPage.Length.ToString(), url2.url, MsgColorEnd));
                }

                switch (url2.typ)
                {
                    case decodeStr:
                        DecoderMetar(url2, htmlPage);
                        break;
                    case xmlStr:
                        // XmlMetar(url2, htmlPage);
                        break;

                    case txtStr:
                        TxtMetar(url2, htmlPage);
                        break;

                    case htmlStr:
                        TxtMetar(url2, HtmlUtil.HtmlToText(htmlPage));
                        break;

                    case arcStr:
                        ArchiveMetar(url2, htmlPage);
                        break;
                }

                url2.dateTime = DateTime.Now;
                url2.threadId = ThreadIdDone;
              
                lock (urlList)
                {
                    urlIdx++;
                    more = (urlIdx < urlList.Count && --numUrl > 0);
                }
            }
        }

        private void GetWebResponses()
        {
            // threadLoadMetarsAbort = false;

            lock (urlList)
            {
                urlList.Clear();
                List<string> icaoList = new List<string>();

                // Process station URL's
                foreach (ListViewItem staItem in stationListView.CheckedItems)
                {
                    Station station = new Station(staItem);
                    string icao = station.icao;
                    icaoList.Add(icao);

                    foreach (ListViewItem urlItem in urlView.CheckedItems)
                    {
                        string urlTyp = urlItem.SubItems[0].Text;
                        string urlFmt = urlItem.SubItems[1].Text;
                        switch (urlTyp)
                        {
                            case arcStr:
                                break;
                            default:
                                string urlStr = string.Format(urlFmt, icao);
                                Url2 url2 = new Url2(urlStr, urlTyp, station);
                                urlList.Add(url2);
                                break;
                        }
                    }
                }

                // Process any hourly archive URL's
                // Keep track of previous archive queries and only issue query if new hour or
                // new station.
                // TODO - when user clears/deletes stations, make similar change to archive station list.

                List<int> archiveDay = new List<int>();
                List<string> archiveICAO = new List<string>();

                foreach (ListViewItem urlItem in urlView.CheckedItems)
                {
                    if (urlItem.SubItems[0].Text == arcStr)
                    {
                        int day = DateTime.UtcNow.Day;
                        bool anyQuery = false;
                        if (archiveDay.Find(delegate(int t) { return day == t; }) == day)
                        {
                            foreach (string icao in icaoList)
                            {
                                anyQuery = archiveICAO.Find(delegate(string s) { return icao == s; }) != icao;
                                if (anyQuery)
                                {
                                    archiveICAO.Add(icao);
                                    break;
                                }
                            }
                        }

                        archiveDay.Add(day);
                        if (anyQuery)
                        {
                            string urlTyp = urlItem.SubItems[0].Text;
                            string urlFmt = urlItem.SubItems[1].Text;

                            if (urlFmt.Contains("(HH)") && urlTyp == arcStr)
                            {
                                urlFmt = urlFmt.Replace("(HH)", "{0:00}");
                                Station noStation = new Station();
                                for (int hourZ = 0; hourZ < 24; hourZ++)
                                {
                                    string urlHH = string.Format(urlFmt, hourZ);
                                    Url2 url2 = new Url2(urlHH, urlTyp, noStation);
                                    urlList.Add(url2);
                                }
                            }
                        }
                    }
                } 
            }

            StartLoadThread(0, urlList.Count);
        }

        #endregion

        #region ==== Right Side Panel UI
        #region ==== Master Station View Controls
        private void mstatnListView_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Space:
                    CopyMstToSta();
                    break;
            }
        }
        
        private void addMst2StaBtn_Click(object sender, EventArgs e)
        {
            CopyMstToSta();
        }

        private void delMstBtn_Click(object sender, EventArgs e)
        {
            DeleteListView(this.mstatnListView, "Okay to clear List?", "Master Station List");
        }

        private void FilterMst(string pattern, int subIdx)
        {
            if (pattern.Length == 0)
                return;

            try
            {
                ListViewColumnSorter sorter = mstatnListView.ListViewItemSorter as ListViewColumnSorter;
                sorter.SortColumn1 = 0;

                // Regex regex = new Regex(pattern, RegexOptions.IgnoreCase);
                if (mstatnListView.CheckedItems.Count > 0)
                {
                    foreach (ListViewItem mstItem in mstatnListView.CheckedItems)
                    {
                        string text = mstItem.SubItems[subIdx].Text;
                        mstItem.Checked = (text.Length > 0) && 
                                text.StartsWith(pattern, StringComparison.OrdinalIgnoreCase);
                    }
                }
                else
                {
                    foreach (ListViewItem mstItem in mstatnListView.Items)
                    {
                        string text = mstItem.SubItems[subIdx].Text;
                        mstItem.Checked  = (text.Length > 0) &&
                                text.StartsWith(pattern, StringComparison.OrdinalIgnoreCase);
                    }
                }

                sorter.SortColumn1 = 0;
                sorter.SortCheckedState = true;
                sorter.Order1 = SortOrder.Ascending;
                mstatnListView.Sort();
                sorter.SortCheckedState = false;
                filterResetBtn.Enabled = (mstatnListView.CheckedItems.Count > 0);
                mstStatus.Text = string.Format("Stations: {0:#,##0}  Selected: {1:#,##0}",
                        mstatnListView.Items.Count, mstatnListView.CheckedItems.Count);
            }
            catch
            {
            }
        }

        private void defaultMstBtn_Click(object sender, EventArgs e)
        {
            Assembly a = Assembly.GetExecutingAssembly();
            Stream zipStream = a.GetManifestResourceStream("WebMetar_NS.mststatn2010.dat.gz");
            GZipStream decompressStream = new GZipStream(zipStream, CompressionMode.Decompress, false);
            TextReader txtReader = new StreamReader(decompressStream);
            LoadMasterStation(txtReader);
        }

        private void filterCountry_TextChanged(object sender, EventArgs e)
        {
            FilterMst(filterCountry.Text, 0);
        }

        private void filterState_TextChanged(object sender, EventArgs e)
        {
            FilterMst(filterState.Text, 1);
        }

        private void filterPlace_TextChanged(object sender, EventArgs e)
        {
            FilterMst(filterPlace.Text, 2);
        }

        private void filterIcao_TextChanged(object sender, EventArgs e)
        {
            FilterMst(filterIcao.Text, 3);
        }

        private void filterResetBtn_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem mstItemCk in mstatnListView.CheckedItems)
            {
                mstItemCk.Checked = false;
            }
            filterResetBtn.Enabled = false;
        }

        private void filter_Click(object sender, EventArgs e)
        {
            TextBox textBox = (TextBox)sender;
            textBox.SelectAll();
        }

        #endregion

        #region ==== Log View Controls
        private void clearLogBtn_Click(object sender, EventArgs e)
        {
            DeleteListView(this.logView, "Okay to clear Log?", "Log");
        }

        private void saveLogBtn_Click(object sender, EventArgs e)
        {
            ListViewExt.ExportDialog(logView, "Save Log");
        }

        private void printLogBtn_Click(object sender, EventArgs e)
        {
            printListView.Print(this, logView, PrintTitle() + "Log");
        }

        private void logView_SelectedIndexChanged(object sender, EventArgs e)
        {
            ShowToolTip(sender);
        }
        #endregion

        #region ==== Web response View Controls
        private void webResponseView_SelectedIndexChanged(object sender, EventArgs e)
        {
            ShowToolTip(sender);
        }

        private void clearWebBtn_Click(object sender, EventArgs e)
        {
            DeleteListView(this.webResponseView, "Okay to clear Web Response?", "Web Response");
        }

        private void saveWebBtn_Click(object sender, EventArgs e)
        {
            ListViewExt.ExportDialog(this.webResponseView, "Save Web Reponse");
        }

        private void printWebBtn_Click(object sender, EventArgs e)
        {
            printListView.Print(this, webResponseView, PrintTitle() + "Web Responses");
        }
        #endregion

        #region ==== URL view controls
        private void urlAddBtn_Click(object sender, EventArgs e)
        {
            this.urlView.Items.Add("Raw").SubItems.Add(urlRaw);
        }

        private void urlDelBtn_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem urlItem in urlView.SelectedItems)
            {
                urlView.Items.Remove(urlItem);
            }
        }
        #endregion
     
        #region ==== Unit View controls
        bool okayUpdateUnitView = true;
        const string onStr = "a";
        const string offStr = "q";
        string graphParameter = "Temperature";

        // Mark Graph column (must be numeric value);
        private ColumnObj SetUnitGraph(string parameter)
        {
            ColumnObj columnObj = null;
            foreach (ListViewItem unitItem in this.unitView.Items)
            {
                if (unitItem.SubItems[0].Text == parameter)
                {
                    columnObj = (ColumnObj)unitItem.Tag;
                    if (columnObj.m_wxbase.IsNumber)
                    {
                        unitItem.SubItems[3].Text = onStr;
                    }
                    else
                    {
                        columnObj = null;
                    }
                }
            }

            if (columnObj != null)
            {
                foreach (ListViewItem unitItem in this.unitView.Items)
                {
                    if (unitItem.SubItems[0].Text != parameter)
                        unitItem.SubItems[3].Text = offStr;
                }
            }

            return (columnObj != null) ? columnObj : settings.graphColumn;
        }

        private void UpdateUnitView()
        {
            if (okayUpdateUnitView)
            {
                okayUpdateUnitView = false;
                this.unitView.Items.Clear();

                settings.columnOrder.m_colOrdInfoList.Sort(new ColumnOrder.DspCompare());
                foreach (ColumnOrder.ColOrdInfo ci in settings.columnOrder.m_colOrdInfoList)
                {
                    ListViewItem unitItem = this.unitView.Items.Add(ci.text);
                    unitItem.UseItemStyleForSubItems = false;

                    unitItem.Checked = (ci.width > 0);

                    ColumnObj columnObj = (ColumnObj)this.tableDataView.Columns[ci.index].Tag;
                    if (columnObj != null)
                    {
                        unitItem.SubItems.Add(columnObj.m_wxbase.UnitText);
                        unitItem.SubItems.Add(columnObj.m_wxbase.Format);
                        unitItem.SubItems.Add("");  // graph on/off

                        float fontSize = this.Font.Size * 1.5f;
                        unitItem.SubItems[3].Font = new Font("Marlett", fontSize);
                        unitItem.SubItems[3].Text = (ci.text == graphParameter) ? onStr : offStr;
                    }

                    unitItem.Tag = columnObj;
                    okayUpdateUnitView = true;
                }
                // settings.columnOrder.m_colOrdInfoList.Sort(new ColumnOrder.IdxCompare());
            }
        }

        private void FillUnitView()
        {
            this.tableDataView.Columns[0].Tag = new ColumnObj(0, settings.presentation.country);
            this.tableDataView.Columns[1].Tag = new ColumnObj(1, settings.presentation.state);
            this.tableDataView.Columns[2].Tag = new ColumnObj(2, settings.presentation.place);
            this.tableDataView.Columns[3].Tag = new ColumnObj(3, settings.presentation.icao);
            this.tableDataView.Columns[4].Tag = new ColumnObj(4, settings.presentation.time);
            this.tableDataView.Columns[5].Tag = new ColumnObj(5, settings.presentation.temperature);
            this.tableDataView.Columns[6].Tag = new ColumnObj(6, settings.presentation.dewpoint);
            this.tableDataView.Columns[7].Tag = new ColumnObj(7, settings.presentation.wind.Direction);
            this.tableDataView.Columns[8].Tag = new ColumnObj(8, settings.presentation.wind.Speed);
            this.tableDataView.Columns[9].Tag = new ColumnObj(9, settings.presentation.wind.Gust);
            this.tableDataView.Columns[10].Tag = new ColumnObj(10, settings.presentation.pressure);
            this.tableDataView.Columns[11].Tag = new ColumnObj(11, settings.presentation.precipitation);
            this.tableDataView.Columns[12].Tag = new ColumnObj(12, settings.presentation.clouds);
            this.tableDataView.Columns[13].Tag = new ColumnObj(13, settings.presentation.visibility);
            this.tableDataView.Columns[14].Tag = new ColumnObj(14, settings.presentation.weather);
            this.tableDataView.Columns[15].Tag = new ColumnObj(15, settings.presentation.pulled);
            this.tableDataView.Columns[16].Tag = new ColumnObj(16, settings.presentation.age);
            this.tableDataView.Columns[17].Tag = new ColumnObj(17, settings.presentation.latlon);
            this.tableDataView.Columns[18].Tag = new ColumnObj(18, settings.presentation.latlon);
        }

        private void unitView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            if (okayUpdateUnitView)
            {
                if (e.Item.Tag != null)
                {
                    int idx = ((ColumnObj)e.Item.Tag).m_colIdx;
                    int w = (e.Item.Checked) ? 60 : 0;
                    this.tableDataView.Columns[idx].Width = w;
                    settings.columnOrder.m_colOrdInfoList[idx].width = w;
                }
            }
        }

        private void upUnitBtn_Click(object sender, EventArgs e)
        {
            if (unitView.SelectedItems.Count != 0)
            {
                ListViewItem item = unitView.SelectedItems[0];
                if (item.Index > 0)
                {
                    int idx = item.Index - 1;
                    item = unitView.Items[idx];
                    unitView.Items.RemoveAt(idx);
                    unitView.Items.Insert(idx + unitView.SelectedItems.Count, item);
                    UpdateTableDataColumns();
                }
            }
        }

        private void downUnitBtn_Click(object sender, EventArgs e)
        {
            int cnt = unitView.SelectedItems.Count;
            if (cnt != 0)
            {
                ListViewItem item = unitView.SelectedItems[cnt - 1];
                if (item.Index < unitView.Items.Count - 1)
                {
                    int idx = item.Index + 1;
                    item = unitView.Items[idx];
                    unitView.Items.RemoveAt(idx);
                    unitView.Items.Insert(idx - unitView.SelectedItems.Count, item);
                    UpdateTableDataColumns();
                }
            }
        }
        #endregion

        #region ==== Detail Table UI
        /// <summary>
        /// Adjust sub table zones
        /// </summary>
        private void TableSize(int panelIdx, float percent)
        {
            this.detailTblPanel.Visible = false;
            if (detailTblPanel.RowStyles[panelIdx].Height >= percent)
                percent = Math.Min(100, detailTblPanel.RowStyles[panelIdx].Height + percent / 2);
            float avgPercent = (100 - percent) / (detailTblPanel.RowStyles.Count - 1);
            for (int scanIdx = 0; scanIdx < detailTblPanel.RowStyles.Count; scanIdx++)
            {
                RowStyle rowStyle = detailTblPanel.RowStyles[scanIdx];
                rowStyle.Height = (scanIdx == panelIdx) ? percent : avgPercent;
            }
            this.detailTblPanel.Visible = true;
        }

        /// <summary>
        /// Set active table button colors
        /// </summary>
        Color onBtnColor = Color.Aqua;
        private void ToolStripOn(ToolStrip toolStrip, ToolStripButton button)
        {
            foreach (ToolStripItem tsItem in toolStrip.Items)
            {
                if (tsItem is ToolStripButton)
                {
                    ToolStripButton tsButton = (ToolStripButton)tsItem;
                    if (tsButton != button)
                    {
                        tsButton.BackColor = Color.Transparent; //  detailToolStrip.BackColor;
                    }
                }
            }

            button.BackColor = onBtnColor;
        }

        private void detailTextBtn_Click(object sender, EventArgs e)
        {
            TableSize(0, 100);
            ToolStripOn(detailToolStrip, detailTextBtn);
        }

        private void detailGraphBtn_Click(object sender, EventArgs e)
        {
            TableSize(1, 100);
            ToolStripOn(detailToolStrip, detailGraphBtn);
        }

        private void detailTimeBtn_Click(object sender, EventArgs e)
        {
            TableSize(2, 100);
            ToolStripOn(detailToolStrip, detailTimeBtn);
        }

        private void detailTblAllBtn_Click(object sender, EventArgs e)
        {
            TableSize(0, 30);
            TableSize(1, 30);
            ToolStripOn(detailToolStrip, detailTblAllBtn);
        }

      
        Point startPos = Point.Empty;
        int movePanel = -1;

        /// <summary>
        /// Resize panels via vertical drag of sash dividers.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            Point mouseLocation = System.Windows.Forms.Control.MousePosition;
            Point panelPos = detailTblPanel.PointToClient(mouseLocation);

            if (e.Button == MouseButtons.Left && detailTblPanel.DisplayRectangle.Contains(panelPos))
            {
                const int minPanel = 0;
                int tableHeight = detailTblPanel.Height - 8;

                if (startPos == Point.Empty)
                {
                    startPos = mouseLocation;
                    movePanel = 0;
                    int partHeight = 0;
                    while (panelPos.Y > partHeight && movePanel < detailTblPanel.RowStyles.Count)
                    {
                        partHeight += (int)(detailTblPanel.RowStyles[movePanel].Height * detailTblPanel.Height / 100);
                        movePanel++;
                    }

                    movePanel = Math.Min(movePanel, detailTblPanel.RowStyles.Count - 1);
                }
                else
                {
                    int yDelta = mouseLocation.Y - startPos.Y;
                    startPos = mouseLocation;

                    float oldPercent = detailTblPanel.RowStyles[movePanel].Height;
                    int h = (int)(oldPercent / 100 * tableHeight + 0.5);
                    float newPercent = Math.Min(tableHeight, Math.Max(minPanel, h - yDelta)) * 100.0f / tableHeight;

                    detailTblPanel.RowStyles[movePanel].Height = newPercent;

                    float rmPercent = newPercent - oldPercent;
                    for (int panelIdx = 0; rmPercent != 0 && panelIdx < detailTblPanel.RowStyles.Count; panelIdx++)
                    {
                        if (panelIdx != movePanel)
                        {
                            if (rmPercent > 0)
                            {
                                if (detailTblPanel.RowStyles[panelIdx].Height > 0)
                                {
                                    float chg = Math.Min(rmPercent, detailTblPanel.RowStyles[panelIdx].Height);
                                    detailTblPanel.RowStyles[panelIdx].Height -= chg;
                                    rmPercent -= chg;
                                }
                            }
                            else
                            {
                                if (detailTblPanel.RowStyles[panelIdx].Height < 100)
                                {
                                    float chg = Math.Max(rmPercent, detailTblPanel.RowStyles[panelIdx].Height - 100);
                                    detailTblPanel.RowStyles[panelIdx].Height -= chg;
                                    rmPercent -= chg;
                                }
                            }
                        }
                    }
                }
            }
            else
            {
                startPos = Point.Empty;
            }

            base.OnMouseMove(e);
        }

        /// <summary>
        /// Global mouse override above does not work, so capture panel specific mouse move
        /// and call global OnMouseMove method.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tableSplit0_MouseMove(object sender, MouseEventArgs e)
        {
            Cursor.Current = Cursors.HSplit;
            OnMouseMove(e);
        }

        #endregion

        #region ==== Detail Graphs

        private void tabData_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (tabData.SelectedTab == tabDetail)
            {
                UpdateGraph();
            }
            else if (tabData.SelectedTab == tabMap)
            {
                mapRadio_Click(null, EventArgs.Empty);
                PanImage(mapImage, mapOrigin);
            }
        }

        private void barGraphProperty_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            this.barGraph.UpdateInternalSettings();
            this.barGraph.Refresh();
        }

        private void timeGraphProperty_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            this.timeGraph.Refresh();
        }

        private void detailTblRefreshBtn_Click(object sender, EventArgs e)
        {
            UpdateGraph();
        }
      

        PrintIt printIt = new PrintIt();
        private void detailTblPrint_Click(object sender, EventArgs e)
        {
            if (detailTextBtn.BackColor == onBtnColor)
            {
                Rectangle screenRect = new Rectangle(detailTblPanel.Location, detailTblPanel.Size);
                printIt.Print(this, screenRect, PrintTitle());
            }
            else if (detailGraphBtn.BackColor == onBtnColor)
            {
                printIt.Print(this, this.barGraph, PrintTitle());
            }
            else if (detailTimeBtn.BackColor == onBtnColor)
            {
                printIt.Print(this, this.timeGraph, PrintTitle());
            }
            else if (detailTblAllBtn.BackColor == onBtnColor)
            {
                Point screenPt = detailTblPanel.PointToScreen(Point.Empty);
                Rectangle screenRect = new Rectangle(screenPt, detailTblPanel.Size);  
                printIt.Print(this, screenRect, PrintTitle());
            }
        }

        #endregion

        #region ==== Configuration UI

        private void cfgWebFetchSettings_ValueChanged(object sender, EventArgs e)
        {
            decimal seconds = cfgWebFetchSettings.Value;
            if (seconds > 0 && seconds < 36000)
                settings.waitSettings = (int)seconds;
        }

        private void ageColorView_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Delete:
                    foreach (ListViewItem ageItem in ageColorView.SelectedItems)
                    {
                        TimeSpan age = (TimeSpan)ageItem.Tag;
                        settings.config.ageColors.Remove(age);
                        ageColorView.Items.Remove(ageItem);
                    }
                    break;
                case Keys.Insert:
                    {
                        ListViewItem ageItem = ageColorView.Items.Add(
                            TimeSpanToString(TimeSpan.MinValue));
                        ageItem.UseItemStyleForSubItems = false;
                        ageItem.SubItems.Add("").BackColor = Color.Transparent;
                        ageItem.SubItems.Add("").BackColor = Color.Transparent;
                        ageItem.Tag = TimeSpan.MinValue;
                    }
                    break;
            }
        }

        private void cfgSoundView_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Delete:
                    break;
                case Keys.Insert:
                    break;
            }
        }

        private void cfgSoundView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            int soundIdx = e.Item.Index;
            settings.config.sounds[soundIdx].Enabled = e.Item.Checked;
        }

        private void cfgSoundView_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (cfgSoundView.SelectedIndices.Count > 0)
            {
                int soundIdx = cfgSoundView.SelectedIndices[0];
                bool enabled = settings.config.sounds[soundIdx].Enabled;
                settings.config.sounds[soundIdx].Enabled = true;
                PlaySound(settings.config.sounds[soundIdx].action);
                settings.config.sounds[soundIdx].Enabled = enabled;
            }
        }

        #endregion

        #endregion

        #region ==== Left Side Panel UI
        #region ==== Table Metar Data view
        private void tblRefreshBtn_Click(object sender, EventArgs e)
        {
            this.sPlayer.Stream = Properties.Resources.click1;
            this.sPlayer.Play();

            threadLoadMetarsAbort = false;
            UpdateStationList();
            GetWebResponses();
        }

        private void saveTblBtn_Click(object sender, EventArgs e)
        {
            ListViewExt.ExportDialog(tableDataView, "Save METAR data table");
        }

        private void printTblBtn_Click(object sender, EventArgs e)
        {
            printListView.Print(this, this.tableDataView, PrintTitle() + "METAR Data Table");
        }

        private void deleteTblBtn_Click(object sender, EventArgs e)
        {
            string prompt = "Okay to clear METAR table?";
            string title = "METAR Table";

            if (tableDataView.SelectedItems.Count == 0)
            {
                if (MessageBox.Show(prompt, title, MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    for (int idx = tableDataView.Items.Count - 1; idx >= 0; idx--)
                    {
                        ListViewItem metarItem = tableDataView.Items[idx];
                        Metar metar = (Metar)metarItem.Tag;
                        oldMetarList.Remove(metar);
                        tableDataView.Items.Remove(metarItem);
                    }
                }
            }
            else
            {
                foreach (ListViewItem metarItem in tableDataView.SelectedItems)
                {
                    Metar metar = (Metar)metarItem.Tag;
                    oldMetarList.Remove(metar);
                    tableDataView.Items.Remove(metarItem);
                }
            }

            UpdateStationList();
            UpdateTableStatus();
            UpdateGraph();
        }

        private void tableDataView_ColumnReordered(object sender, ColumnReorderedEventArgs e)
        {
            // Mark as reordered and call method in timer because columns have not actually changed yet.
            columnReordered = true;
        }

        private void ColumnReordered()
        {
            columnReordered = false;
            settings.columnOrder.GetViewOrder(tableDataView);
            UpdateUnitView();
        }

        private void UpdateTableDataColumns()
        {
            okayUpdateUnitView = false;
            foreach (ListViewItem unitItem in unitView.Items)
            {
                int idx = ((ColumnObj)unitItem.Tag).m_colIdx;
                if (unitItem.Checked)
                {
                    if (this.tableDataView.Columns[idx].Width < 5)
                    {
                        this.tableDataView.Columns[idx].Width = 60;
                    }
                }
                else
                {
                    this.tableDataView.Columns[idx].Width = 0;
                }

                this.tableDataView.Columns[idx].DisplayIndex = unitItem.Index;
            }
            okayUpdateUnitView = true;
        }

        private void AppendText(string s1, string s2)
        {
            this.detailText.SelectionFont = new Font(this.detailText.Font, FontStyle.Bold);
            this.detailText.SelectionColor = Color.Black;
            this.detailText.SelectedText = s1 + "\t";

            this.detailText.SelectionFont = new Font(this.detailText.Font, FontStyle.Regular);
            this.detailText.SelectionColor = Color.Blue;                                       
            this.detailText.SelectedText = s2 + "\n";
        }

        private void tableDataView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            ListView listView = tableDataView;

            Control c = listView;
            Point p = listView.PointToClient(Control.MousePosition);
            ListViewHitTestInfo hitInfo = listView.HitTest(p);
            if (hitInfo.SubItem != null)
            {
                Metar metar = (Metar)hitInfo.Item.Tag;
                detailText.Clear();

                detailText.SelectedText = metar.metarStr + "\n" + metar.notes;

                if (metar.notes.Length == 0)
                {
                    detailText.Text += "\n\n";
                    AppendText("ICAO:          ", metar.icao );
                    AppendText("TimeUTC:       " , metar.reportTimeUTC.ToString("G"));
                    detailText.SelectedText = "\n";

                    AppendText("Temperature:   " , metar.temperature.TextandUnit);
                    AppendText("Dew Point:     " , metar.dewpoint.TextandUnit);
                    AppendText("Wind Dir:      ", metar.wind.Direction.TextandUnit);
                    AppendText("Wind Speed:    ", metar.wind.Speed.TextandUnit);
                    AppendText("Wind Gust:     ", metar.wind.Gust.TextandUnit);
                    AppendText("Pressure:      ", metar.pressure.TextandUnit);
                    AppendText("Precipitation: ", metar.precipitation.TextandUnit);
                    // AppendText("Humidity:      ", metar.humidity.TextandUnit);

                    foreach (Visibility visibility in metar.visibilityList)
                        AppendText("Visibility:    ", visibility.TextandUnit);

                    foreach (Clouds clouds in metar.cloudsList)
                        AppendText("Clouds:        ", clouds.TextandUnit);

                    foreach (Weather weather in metar.weatherList)
                        AppendText("Weather:       ", weather.TextandUnit);

                    AppendText("Remarks:       ", metar.remarks); 
                    detailText.SelectedText += "\n\n";
                }

                this.tabData.SelectedTab = this.tabDetail;
                UpdateBarGraph();
            }
        }

        private void tableFilter_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Enter:
                    FilterTable(tableFilter.Text);
                    break;
            }
        }

        private void FilterTable(string pattern)
        {
            if (pattern.Length == 0)
                return;

            try
            {
                ListViewColumnSorter sorter = tableDataView.ListViewItemSorter as ListViewColumnSorter;
                sorter.SortColumn2 = sorter.SortColumn1;
                sorter.Order2 = sorter.Order1;
                sorter.Order1 = SortOrder.None;

                Regex regex = new Regex(pattern, RegexOptions.IgnoreCase);
                foreach (ListViewItem tblItem in tableDataView.Items)
                {
                    tblItem.Checked = false;

                    for (int subIdx = 0; subIdx < tblItem.SubItems.Count; subIdx++)
                    {
                        // Only filter viewable columns
                        if (tableDataView.Columns[subIdx].Width > 0)
                        {
                            string text = tblItem.SubItems[subIdx].Text;
                            if (text.Length > 0 && regex.IsMatch(text))
                            {
                                // tblItem.SubItems[6].Text = "Match";
                                tblItem.Checked = true;
                                break;
                            }
                        }
                    }
                }

                sorter.SortColumn1 = 0;
                sorter.SortCheckedState = true;
                // sorter.SortType = ListViewColumnSorter.SortDataType.eAlpha;
                sorter.Order1 = SortOrder.Ascending;

                tableDataView.Sort();
            }
            catch
            {
            }
        }

        private void viewTblGrpMenu_Click(object sender, EventArgs e)
        {
            ToolStripDropDownItem dropItem = (ToolStripDropDownItem)sender;
            ToolStripDropDownMenu dropMenu = dropItem.DropDown.OwnerItem.Owner as ToolStripDropDownMenu;

            foreach (ToolStripDropDownItem item in dropMenu.Items)
            {
                System.Windows.Forms.ToolStripMenuItem itemBtn =
                    (System.Windows.Forms.ToolStripMenuItem)item;
                if (dropItem != item)
                {
                    itemBtn.Checked = false;
                }
            }

            tableGroup.Clear();
            UpdateMetarView();
            // switch (dropItem.Text)
        }

        #endregion

        #region ==== Query/Data Station List Controls
        private void loadStaBtn_Click(object sender, EventArgs e)
        {
            ListViewExt.ImportDialog(stationListView, "Load Station List");
            UpdateStationStatus();
        }

        private void saveStaBtn_Click(object sender, EventArgs e)
        {
            ListViewExt.ExportDialog(stationListView, "Save Station List");
        }

        private void addStaBtn_Click(object sender, EventArgs e)
        {
            ListViewItem staItem = this.stationListView.Items.Add("<country>");
            staItem.SubItems.Add("<state>");
            staItem.SubItems.Add("<place>");
            staItem.SubItems.Add("<icao>");
            UpdateStationStatus();
            this.stationListView.EnsureVisible(staItem.Index);

            // Point p = staItem.Position;
            Point p = staItem.SubItems[3].Bounds.Location;
            ListViewEdit(stationListView, p);
            UpdateStationList(staItem);
            staItem.Checked = true;
        }

        private void UpdateStationList(ListViewItem staItem)
        {
            string icao = staItem.SubItems[3].Text.ToUpper();
            foreach (ListViewItem mstItem in mstatnListView.Items)
            {
                if (mstItem.SubItems[3].Text == icao)
                {
                    staItem.SubItems[0].Text = mstItem.SubItems[0].Text;   // country
                    staItem.SubItems[1].Text = mstItem.SubItems[1].Text;   // state
                    staItem.SubItems[2].Text = mstItem.SubItems[2].Text;   // place
                    staItem.SubItems[3].Text = mstItem.SubItems[3].Text;   // icao
                    staItem.SubItems[4].Text = mstItem.SubItems[4].Text;   // lat
                    staItem.SubItems[5].Text = mstItem.SubItems[5].Text;   // lon
                }
            }
        }

        private void delStaBtn_Click(object sender, EventArgs e)
        {
            DeleteListView(this.stationListView, "Okay to clear list?", "Data Station List");
            UpdateStationStatus();
        }

        private void stationListView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            UpdateStationStatus();
        }

        private void UpdateStationStatus()
        {
            stationStatus.Text = "Stations:" + stationListView.Items.Count.ToString("#,##0") +
                 "   Checked:" + stationListView.CheckedItems.Count.ToString("#,##0");
        }

        private void UpdateStationList()
        {
            // Rebuild station list with checked stations.
            lock (stationListView)
            {
                stationList.Clear();
                foreach (ListViewItem staItem in stationListView.CheckedItems)
                {
                    Station station = new Station(staItem);
                    if (false == stationList.ContainsKey(station.icao))
                    {
                        stationList.Add(station.icao, station);
                    }
                }
            }
        }

        #endregion

        #region ===== Web Browser controls
        private void urlBroText_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Enter:
                    BrowseTo(urlBroText.Text);
                    break;
            }
        }

        private void UpdateUrl()
        {
            this.urlBroText.Text = this.webBrowser.Url.AbsoluteUri;
        }

        private void comboBro_SelectedIndexChanged(object sender, EventArgs e)
        {
            string url = comboBro.SelectedItem.ToString();
            this.urlBroText.Text = url;
            // BrowseTo(url);
        }

        private void BrowseTo(string text)
        {
            this.webBrowser.Navigate("about:Loading page " + text + " ...");
            Application.DoEvents();

            this.webBrowser.Navigate(urlBroText.Text);
            if (false == this.comboBro.Items.Contains(text))
            {
                this.comboBro.Items.Add(text);
            }
        }

        private void backBroBtn_Click(object sender, EventArgs e)
        {
            this.webBrowser.GoBack();
            UpdateUrl();
        }

        private void fwdBroBtn_Click(object sender, EventArgs e)
        {
            this.webBrowser.GoForward();
            UpdateUrl();
        }

        private void webBrowser_DocumentCompleted(object sender, WebBrowserDocumentCompletedEventArgs e)
        {
            if (webBrowser.Document != null
                && webBrowser.Document.Images != null
                && webBrowser.Document.Links != null)
            {
                this.statusBro.Text = string.Format("Image:{0}, Links{1}",
                    webBrowser.Document.Images.Count,
                    webBrowser.Document.Links.Count);
            }
        }
        #endregion

        #region ==== Bookmarks

 
        private void SetBookmark(ListViewItem bmItem)
        {
            Bookmark bookmark = new Bookmark();
            bookmark.stations = new List<Station>();
            bookmark.presentation = settings.presentation;
            bookmark.name = bmItem.SubItems[1].Text;
            bookmark.color = bmItem.SubItems[2].BackColor;
            bookmark.keycode = Keys.None;       // parse subItems[3].Text
            bookmark.description = bmItem.SubItems[4].Text;

            int staCnt = stationListView.CheckedItems.Count;
            if (staCnt > 0)
            {
                foreach (ListViewItem staItem in stationListView.Items)
                {
                    bookmark.stations.Add(new Station(staItem));
                }
            }

            bmItem.Tag = bookmark;
        }

        bool bookmarkDoChecked = true;
        private void bookmarkAddBtn_Click(object sender, EventArgs e)
        {
            if (this.bookmarkView.SelectedItems.Count > 0)
            {
                SetBookmark(this.bookmarkView.SelectedItems[0]);
            }
            else
            {
                bookmarkDoChecked = false;
                // Bookmark fields:
                //  0 = on
                //  1 = name
                //  2 = color
                //  3 = key
                //  4 = desc
                ListViewItem bmItem = this.bookmarkView.Items.Add("");
                bmItem.UseItemStyleForSubItems = false;
                bmItem.SubItems.Add("bookmark" + this.bookmarkView.Items.Count.ToString());
                bmItem.SubItems.Add(" ");
                bmItem.SubItems.Add("no key");
                bmItem.SubItems.Add("need description");
                string desc = string.Empty;
                int staCnt = stationListView.CheckedItems.Count;
                for (int staIdx = 0; staIdx < 5 && staIdx < staCnt; staIdx++)
                {
                    desc += stationListView.CheckedItems[staIdx].SubItems[3].Text + " ";
                }
                bmItem.SubItems[4].Text = desc;
                bmItem.SubItems[2].BackColor = Color.LightGreen;

                SetBookmark(bmItem);
                // ListViewEdit(bookmarkView, bmItem.Position);
                bmItem.Checked = true;
                ActivateBookmark(bmItem);
                bookmarkDoChecked = true;
            }
        }

        private void bookmarkDelBtn_Click(object sender, EventArgs e)
        {
            DeleteListView(this.bookmarkView, "Okay to clear list ?", "Bookmarks");
        }

        private void ActivateBookmark(ListViewItem bmItem)
        {
            if (bookmarkDoChecked)
            {
                bookmarkDoChecked = false;      // prevent recursion

                bmItem.Checked = true;
                Bookmark bookmark = (Bookmark)bmItem.Tag;
                if (bookmark != null)
                {
                    this.bookmarkLabel.Text = bookmark.name;
                    this.bookmarkLabel.ToolTipText = bookmark.description;
                    this.bookmarkLabel.BackColor = bookmark.color;

                    Bitmap image = new Bitmap(8, 8);
                    Graphics g = Graphics.FromImage(image);
                    g.Clear(bookmark.color);
                    this.bookmarkLabel.BackgroundImage = image;

                    this.settings.presentation = bookmark.presentation;
                    UpdateUnitView();

                    settings.stations = bookmark.stations;
                    UpdateStationView();
                }

                foreach (ListViewItem bmItemCk in bookmarkView.CheckedItems)
                {
                    if (bmItemCk != bmItem)
                        bmItemCk.Checked = false;
                }

                bookmarkDoChecked = true;
            }
        }

        private void bookmarkView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            ActivateBookmark(e.Item);
        }

        #endregion
        #endregion

        #region ==== Drag and Drop from Mststatn to StationList

        private void stationListView_DragDrop(object sender, DragEventArgs e)
        {
            string dropStr = string.Empty;
            if (e.Data.GetDataPresent(DataFormats.OemText))
            {
                dropStr = (string)e.Data.GetData(DataFormats.OemText);
            }
            else if (e.Data.GetDataPresent(DataFormats.Text))
            {
                dropStr = (string)e.Data.GetData(DataFormats.Text);
            }

            if (dropStr != string.Empty)
            {
                ListViewItem staItem = this.stationListView.Items.Add("");  // country
                staItem.SubItems.Add("");  // state 
                staItem.SubItems.Add("");  // place
                staItem.SubItems.Add("");  // icao
                staItem.SubItems.Add("");  // lat
                staItem.SubItems.Add("");  // lon

                staItem.SubItems[3].Text = dropStr;
                UpdateStationList(staItem);
                staItem.Checked = true;
            }
            else
                MessageBox.Show("Bad drop");
        }

        private void stationListView_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.Text) || e.Data.GetDataPresent(DataFormats.OemText))
                e.Effect = DragDropEffects.Copy;
            else
                e.Effect = DragDropEffects.None;
        }

        /// <summary>
        /// Start a drag (mouse move with left button down)
        /// </summary>
        private void dragStart_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left /* && e.Clicks == 1 */)
            {
                ListViewItem item = this.mstatnListView.GetItemAt(e.X, e.Y);

                if (item != null)
                {
                    if (DoDragDrop(item.SubItems[3].Text, DragDropEffects.Copy) == DragDropEffects.Copy)
                    {
                    }
                }
            }
        }
        #endregion

        #region ===== Map Viewer Tab

        Bitmap mapImage = global::WebMetar_NS.Properties.Resources.unitedstates;
        PointF mapOrigin = Point.Empty;
        Point mapStart = Point.Empty;
        SizeF mapZoom;
        Size mapBoxBeforeZoom;
        Color mapMarkColor = Color.Red;

        Size mapBoxOffset = new Size (80, 220);

        private void mapBox_Resize(object sender, EventArgs e)
        {
            float resizeW = mapZoom.Width + (mapBox.Width - mapBoxBeforeZoom.Width);
            float resizeH = mapZoom.Height + (mapBox.Height - mapBoxBeforeZoom.Height);

            if (resizeW > 16 && resizeW < mapImage.Width * 2 &&
                resizeW > 16 && resizeH < mapImage.Height * 2)
            {
                mapZoom.Width = resizeW;
                mapZoom.Height = resizeH;
            }

            mapBoxBeforeZoom = mapBox.Size;

            // this.mapBox.BackgroundImage = new Bitmap(mapBox.Width, mapBox.Height);
            // this.mapBox.Image = new Bitmap(mapBox.Width, mapBox.Height);
        }
   
        private void PanImage(Image srcImage, PointF ur)
        {
            Image dstImage = mapBox.BackgroundImage;
            if (dstImage.Size != mapBox.Size)
            {
                dstImage = mapBox.BackgroundImage = new Bitmap(mapBox.Width, mapBox.Height);
                mapBox.Image = new Bitmap(mapBox.Width, mapBox.Height);
            }

            Graphics g = Graphics.FromImage(dstImage);
            g.Clear(Color.Transparent);
            RectangleF clipRect = new RectangleF(ur, mapZoom);
            RectangleF zoomRect = new RectangleF(Point.Empty, mapBox.Size);
            g.InterpolationMode = InterpolationMode.Low;
            g.DrawImage(srcImage, zoomRect, clipRect, GraphicsUnit.Pixel);
            g.Dispose();

            // Draw any makers
            if (mapMarkerList.Count > 0)
            {
                Graphics gTop = Graphics.FromImage(this.mapBox.Image);
                gTop.Clear(Color.Transparent);
                Brush markerBrush = new SolidBrush(this.mapMarkColor);
                float zoomX = (float)mapBox.Width / mapZoom.Width;
                float zoomY = (float)mapBox.Height / mapZoom.Height;
                int markSize = this.markSizeBar.Value;
                int markSzHalf = markSize / 2;

                foreach (MapMarker mapMarker in mapMarkerList)
                {
                    float x = (int)(mapMarker.m_point.X - ur.X) * zoomX - markSzHalf;
                    float y = (int)(mapMarker.m_point.Y - ur.Y) * zoomY - markSzHalf;
                    mapMarker.m_screen = new PointF(x, y);
                    gTop.FillEllipse(markerBrush, x, y, markSize, markSize);
                }
                gTop.Dispose();
            }

            this.mapBox.Refresh();
        }

        private void markColorBtn_Click(object sender, EventArgs e)
        {
            this.colorDialog.Color = mapMarkColor;
            if (sender == null || this.colorDialog.ShowDialog() == DialogResult.OK)
            {
                mapMarkColor = this.colorDialog.Color;

                Rectangle imgRect = new Rectangle(Point.Empty, markColorBtn.Size);
                Graphics g = Graphics.FromImage(markColorBtn.Image);
                g.FillRectangle(new SolidBrush(mapMarkColor), imgRect);
                g.DrawRectangle(Pens.Black, imgRect);
                imgRect.Inflate(-1, -1);
                g.DrawRectangle(Pens.White, imgRect);

                PanImage(mapImage, mapOrigin);
            }
        }

        private void markSizeBar_Scroll(object sender, EventArgs e)
        {
            PanImage(mapImage, mapOrigin);
        }

        private void zoomBar_Scroll(object sender, EventArgs e)
        {
            float zoomLevel = zoomBar.Value / 100.0f;
            float aspect = (float)mapBox.Height / mapBox.Width;

            float zoomW = mapImage.Width * zoomLevel;
            float zoomH = zoomW * aspect;

            mapOrigin.X += (mapZoom.Width - zoomW) / 2;
            mapOrigin.Y += (mapZoom.Height - zoomH) / 2;

            mapZoom.Width = zoomW;
            mapZoom.Height = zoomH;

            PanImage(mapImage, mapOrigin);
        }

       
        private void mapBox_MouseMove(object sender, MouseEventArgs e)
        {
            int dX = e.X - mapStart.X;
            int dY = e.Y - mapStart.Y;

            // System.Diagnostics.Debug.WriteLine(dX.ToString() + ", " + dY.ToString());
            if (e.Button == MouseButtons.Left)
            {
                // ---- Pan
                float zoomLevel = (float)mapZoom.Width / mapImage.Width;
                mapOrigin.X -= dX * zoomLevel;
                mapOrigin.Y -= dY * zoomLevel;

                PanImage(mapImage, mapOrigin);
            }
#if false
            else if (e.Button == MouseButtons.Right)
            {
                // ---- Zoom
                int zoomXY = (Math.Abs(dX) > Math.Abs(dY)) ? dX : dY;
                float aspect = (float)mapBox.Height / mapBox.Width;

                float zoomW = mapZoom.Width + zoomXY;
                float zoomH = zoomW * aspect;

                if (zoomW > 16 && zoomW < mapImage.Width*2 &&
                    zoomH > 16 && zoomH < mapImage.Height*2)
                {
                    mapZoom.Width = zoomW;
                    mapZoom.Height = zoomH;

                    // Adjust origin to keep same image center
                    float shiftXY = zoomXY / 2f;
                    mapOrigin.X -= shiftXY;  
                    mapOrigin.Y -= shiftXY; 
                }

                PanImage(mapImage, mapOrigin);
            }
#endif
            else
            {
                // mouse hover
                Point p = e.Location;
                foreach (MapMarker mapMarker in mapMarkerList)
                {
                    if (Math.Abs(mapMarker.m_screen.X - p.X) < 10 &&
                        Math.Abs(mapMarker.m_screen.Y - p.Y) < 10)
                    {
                        toolTip.Show(mapMarker.m_station.icao + " " + mapMarker.m_station.place, this.mapBox, 
                            (int)mapMarker.m_screen.X+10, (int)mapMarker.m_screen.Y+10, 3000);
                        mapStatus.Text = mapMarker.m_station.icao + " " + mapMarker.m_station.place + " " +
                            mapMarker.m_station.longitude.Text + " " +
                            mapMarker.m_station.latitude.Text;

                    }
                }
            }

            mapStart = e.Location;
        }

        class MapMarker
        {
            public MapMarker(PointF point, Station station)
            {
                m_point = point;
                m_screen = Point.Empty;
                m_station = station;
            }

            public PointF m_point;
            public PointF m_screen;
            public Station m_station;
        }

        List<MapMarker> mapMarkerList = new List<MapMarker>();

        float mapMinXdeg = -125.0f; // -124.7314f;
        float mapMinYdeg = 25.0f;    // 24.3258f;
        float mapMaxXdeg = -66.8f;   // -67.0085f;
        float mapMaxYdeg = 49.7f;    // 50.0405f;
 
        private PointF GetPointFor(Station station)
        {
            if (station.latitude.IsPopulated == false)
            {
                foreach (ListViewItem mstItem in mstatnListView.Items)
                {
                    if (mstItem.SubItems[3].Text == station.icao)
                    {
                        station.FillItem(mstItem);
                        break;
                    }
                }
            }

            if (station.latitude.IsPopulated == false ||
                station.longitude.IsPopulated == false)
            {
                return PointF.Empty;
            }

            float latitude = (float)station.latitude.Degrees;
            float longitude = (float)station.longitude.Degrees;

            // Todo - navigate to image coordinates.
            // name, unitedstates.png
            // min, -124.7314, 24.3258
            // max, -67.0085, 50.0405
            float xRes = (mapMaxXdeg - mapMinXdeg) / mapImage.Width;
            float yRes = (mapMaxYdeg - mapMinYdeg) / mapImage.Height;
            float x = (longitude - mapMinXdeg) / xRes;
            float y = (mapMaxYdeg - latitude) / yRes;

            return new PointF(x, y);
        }

        
        private void mapRadio_Click(object sender, EventArgs e)
        {
            mapMinXdeg = float.Parse(mapMinXBox.Text);
            mapMinYdeg = float.Parse(mapMinYBox.Text);
            mapMaxXdeg = float.Parse(mapMaxXBox.Text);
            mapMaxYdeg = float.Parse(mapMaxYBox.Text);

            mapMarkerList.Clear();
            if (mapRequestCkdRadio.Checked)
            {
                foreach (ListViewItem staItem in stationListView.CheckedItems)
                {
                    Station station = new Station(staItem);
                    PointF icaoPt = GetPointFor(station);
                    MapMarker mapMarker = new MapMarker(icaoPt, station);
                    mapMarkerList.Add(mapMarker);
                }
            }
            else if (mapRequestRadio.Checked)
            {
                foreach (ListViewItem staItem in stationListView.Items)
                {
                    Station station = new Station(staItem);
                    PointF icaoPt = GetPointFor(station);
                    MapMarker mapMarker = new MapMarker(icaoPt, station);
                    mapMarkerList.Add(mapMarker);
                }
            }
            else if (mapMasterCkdRadio.Checked)
            {
                foreach (ListViewItem mstItem in mstatnListView.CheckedItems)
                {
                    Station station = new Station(mstItem);
                    PointF icaoPt = GetPointFor(station);
                    MapMarker mapMarker = new MapMarker(icaoPt, station);
                    mapMarkerList.Add(mapMarker);
                }
            }
            else // all master
            {
                foreach (ListViewItem mstItem in mstatnListView.Items)
                {
                    Station station = new Station(mstItem);
                    PointF icaoPt = GetPointFor(station);
                    MapMarker mapMarker = new MapMarker(icaoPt, station);
                    mapMarkerList.Add(mapMarker);
                }
            }

#if false
            // Calc bounding box.
            float minX = float.MaxValue;
            float minY = float.MaxValue;
            float maxX = float.MinValue;
            float maxY = float.MinValue;

            foreach (MapMarker mapMarker in mapMarkerList)
            {
                minX = Math.Min(minX, mapMarker.m_point.X);
                minY = Math.Min(minY, mapMarker.m_point.Y);
                maxX = Math.Max(maxX, mapMarker.m_point.X);
                maxY = Math.Max(maxY, mapMarker.m_point.Y);
            }

            // Soften zoom - works well if region is too small.
            float dx = mapOrigin.X - (mapOrigin.X + minX) / 2;
            mapOrigin.X += dx;
            mapOrigin.Y += dx;
            mapZoom = new SizeF((mapZoom.Width + maxX - minX)/2, (mapZoom.Height + maxY - minY)/2);
            zoomBar.Value = Math.Max(5, (int)(mapZoom.Width * 100 / mapImage.Width));
#endif

            PanImage(mapImage, mapOrigin);
        }

        #endregion

        #region ==== Notify Icon

        private void notifyIcon_DoubleClick(object sender, EventArgs e)
        {
            this.Show();
        }

        private void notifyIconToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.ShowInTaskbar = !notifyIconToolStripMenuItem.Checked;
            this.notifyIcon.Visible = notifyIconToolStripMenuItem.Checked; 
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Show();
            this.WindowState = FormWindowState.Normal;
        }

        private void pauseToolStripMenuItem_Click(object sender, EventArgs e)
        {
            runningMenuCk_Click(sender, e);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            notifyIcon.Visible = false;
            this.Close();
        }
        #endregion

        private void helpMenu_Click(object sender, EventArgs e)
        {
            HelpDialog helpDialog = new HelpDialog();
            helpDialog.Show();
        }

        private void vSettingsScrollBar_Scroll(object sender, ScrollEventArgs e)
        {
            int scroll = -e.NewValue * 500 / vSettingsScrollBar.Maximum;
            settings1Grp.Location = new Point(settings1Grp.Location.X, 40 + scroll);
            settings2Grp.Location = new Point(settings2Grp.Location.X, 100 + scroll);
            settings3Grp.Location = new Point(settings3Grp.Location.X, 340 + scroll);
            settings4Grp.Location = new Point(settings4Grp.Location.X, 548 + scroll);
        }

        private void mapBtn_Click(object sender, EventArgs e)
        {
            if (mapCombo.Visible)
            {
                mapBtn.Text = "XY:";
                mapCombo.Visible = false;
                mapMinXBox.Visible = mapMinYBox.Visible = mapMaxXBox.Visible = mapMaxYBox.Visible = true;
            }
            else
            {
                mapBtn.Text = "Maps:";
                mapCombo.Visible = true;
                mapMinXBox.Visible = mapMinYBox.Visible = mapMaxXBox.Visible = mapMaxYBox.Visible = false;
            }
        }

        const string internalMapStr = "internal";
        const string unitedStates48Str = "United States48";
        const string worldStr = "World";
        private void mapCombo_SelectedIndexChanged(object sender, EventArgs e)
        {
            foreach (ListViewItem mapItem in mapListView.Items)
            {
                if (mapItem.Text == mapCombo.Text)
                {
                    if (mapItem.SubItems[1].Text == internalMapStr)
                    {
                        switch (mapItem.Text)
                        {
                            case unitedStates48Str:
                                mapImage = global::WebMetar_NS.Properties.Resources.unitedstates;
                                mapMinXdeg = -125.0f; // -124.7314f;
                                mapMinYdeg = 25.0f;    // 24.3258f;
                                mapMaxXdeg = -66.8f;   // -67.0085f;
                                mapMaxYdeg = 49.7f;    // 50.0405f;
                                break;
                            case worldStr:
                                mapImage = global::WebMetar_NS.Properties.Resources.world;
                                mapMinXdeg = -180f;
                                mapMinYdeg = -90f;
                                mapMaxXdeg = 180f;
                                mapMaxYdeg = 84f;
                                break;
                        }

                        this.mapMinXBox.Text = mapMinXdeg.ToString();
                        this.mapMinYBox.Text = mapMinYdeg.ToString();
                        this.mapMaxXBox.Text = mapMaxXdeg.ToString();
                        this.mapMaxYBox.Text = mapMaxYdeg.ToString();
                    }
                    else
                    {
                        try
                        {
                            StreamReader inStream = new StreamReader(mapItem.SubItems[1].Text);
                            // mapImage = Bitmap.FromStream(inStream);
                            // TODO - parse bound box
                        }
                        catch (Exception ex)
                        {
                            MessageBox.Show(ex.Message, mapItem.SubItems[1].Text);
                        }
                    }

                    mapRadio_Click(null, EventArgs.Empty);
                    PanImage(mapImage, mapOrigin);
                    return;
                }
            }
        }

 
    }
}
