﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;


namespace WebMetar_NS
{
    /// <summary>
    /// METAR class - decode and hold METAR parameters.
    /// 
    /// Example input string:
    /// KMWN 180355Z 17014KT 1/16SM FZFG VV001 M06/M06 RMK VRY LGT ICG
    /// KBED 180356Z AUTO 07012KT 5SM RA BR BKN009 OVC013 02/01 A2986 RMK AO2 SLP126 P0013 T00170006
    /// 
    /// Credits:
    ///   METAR decoder based off of work by rantanplanisback http://www.codeproject.com/KB/cs/weathernotify.aspx
    ///   
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    class Metar
    {
        public string metarStr = string.Empty;
        public string notes = string.Empty;
      
        // decoded metar components
        public string icao = string.Empty;
        public string type = string.Empty;          // METAR or SPECI
        public string remarks = string.Empty;
        public string reportMod = string.Empty;     // Report Modifier: AUTO, COR, CCx or RRx
        public DateTime reportTimeUTC;

        public Temperature temperature = null;
        public Temperature dewpoint = null;
        public Humidity humidity = null;
        public Pressure pressure = null;
        public Precipitation precipitation = null;
        public Wind wind = null;

        const int MaxCloudGroup = 3;
        const int MaxVisibilityGroup = 3;
        const int MaxWeatherGroup = 3;
        const int MaxRunwayGroup = 4;

        public List<Visibility> visibilityList = null;
        public List<Weather> weatherList = null;
        public List<Clouds> cloudsList = null;
        public List<Runway> runwayList = null;

        public Metar()
        {
            // Decode(_metarStr, DateTime.UtcNow);
            ClearFields();
        }

        public void ClearFields()
        {
            this.reportTimeUTC = DateTime.MinValue;
            this.metarStr = string.Empty;
            this.type = string.Empty;
            this.icao = string.Empty;

            this.remarks = string.Empty;

            this.wind = new Wind();
            this.runwayList = new List<Runway>();
            this.visibilityList = new List<Visibility>();
            this.weatherList = new List<Weather>();
            this.cloudsList = new List<Clouds>();

            this.temperature = new Temperature();
            this.dewpoint = new Temperature();
            this.pressure = new Pressure();
            this.humidity = new Humidity();
            this.precipitation = new Precipitation();
        }

        public WxBase GetParameter(string parmName)
        {
            try
            {
                switch (parmName)
                {
                    case "Temperature": return temperature;
                    case "Dew Point": return dewpoint;
                    case "Humidity": return humidity;
                    case "Pressure": return pressure;
                    case "Precipitation": return precipitation;
                    case "Precip": return precipitation;
                    case "Wind Speed": return wind.Speed;
                    case "Wind Gust": return wind.Gust;
                    case "WindDir": return wind.Direction;

                    case "Visibility": return visibilityList[0];
                    case "Cloud": return cloudsList[0];
                    case "Weather": return weatherList[0];

                    case "Age": return new WxTimeSpan();        // ToDo
                }
            }
            catch { }

            return null;
        }

        static double ToDouble(string s)
        {
            if (s == null || s.Length == 0)
                return 0;
            try
            {
                return double.Parse(s);
            }
            catch
            {
                return double.NaN;
            }
        }
        static int ToInt(string s)
        {
            try
            {
                return int.Parse(s);
            }
            catch
            {
                return 0;
            }
        }

        //TFFR 281000Z 00000KT 9999 FEW020 19/18 Q1011 NOSIG
        public bool Decode(string _metarStr, DateTime refTimeUTC)
        {
            metarStr = _metarStr;

            int month = refTimeUTC.Month;
            int year = refTimeUTC.Year;

            /* initialization */
            this.remarks = "";

            RegCompare remarkPat = new RegCompare("(RMK|TEMPO|BECMG|INTER)");
            RegCompare icaoPat = new RegCompare("^[A-Z][A-Z0-9]{3}$");
            RegCompare datePat = new RegCompare("([0-9]{2})([0-9]{2})([0-9]{2})Z");
            RegCompare modPat = new RegCompare("(AUTO|COR|RTD|CC[A-Z]|RR[A-Z])");
            RegCompare windSpdPat = new RegCompare("([0-9]{3}|VRB)([0-9]{2,3})G?([0-9]{2,3})?(KT|MPS|KMH)");
            RegCompare windDirPat = new RegCompare("^([0-9]{3})V([0-9]{3})$");
            RegCompare visPat = new RegCompare("^([0-9]{4})([NS]?[EW]?)$");
            RegCompare temDewPat = new RegCompare("^(M?[0-9]{2})/(M?[0-9]{2})?$");

            RegCompare visDigPat = new RegCompare("^[0-9]$");
            RegCompare visDisPat = new RegCompare("^M?(([0-9]?)[ ]?([0-9])(/?)([0-9]*))SM$");
            string visMileStr = "";

            RegCompare runwayPat = new RegCompare(
                "^R([0-9]{2})([RLC]?)/([MP]?)([0-9]{4})([DNU]?)V?(P?)([0-9]{4})?([DNU]?)$");

            RegCompare wxPat = new RegCompare("^(VC)?" + // Proximity
                    "(-|\\+)?" +                         // Intensity
                    "(MI|PR|BC|DR|BL|SH|TS|FZ)?" +       // Descriptor
                    "((DZ|RA|SN|SG|IC|PL|GR|GS|UP)+)?" + // Precipitation
                    "(BR|FG|FU|VA|DU|SA|HZ|PY)?" +       // Obscuration
                    "(PO|SQ|FC|SS)?$");                  // Other

            RegCompare cloudPat = new RegCompare("^(VV|FEW|SCT|BKN|OVC)([0-9]{3}|///)(CB|TCU)?$");
            RegCompare pressApat = new RegCompare("A([0-9]{4})");
            RegCompare pressQpat = new RegCompare("Q([0-9]{4})");



            string[] parts = this.metarStr.Split(' ');
            int partIdx = 0;
            foreach (string part in parts)
            {
                partIdx++;

                if (remarkPat.IsMatch(part))
                {
                    // "RMK|TEMPO|BECMG|INTER" 
                    // Rest of METAR is either a remark or temporary information.
                    this.remarks = this.metarStr.Substring(this.metarStr.IndexOf(part));
                    string[] remParts = new string[parts.Length - partIdx];
                    Array.Copy(parts, partIdx, remParts, 0, parts.Length - partIdx);
                    DecodeRemarks(remParts);
                }
                else if (part == "METAR")
                {
                    this.type = "METAR";    // Type of report is METAR
                }
                else if (part == "SPECI")
                {
                    this.type = "SPECI";    // Type of report os SPECIal
                }
                else if (icaoPat.IsMatch(part) && (this.icao == ""))
                {
                    this.icao = part;       // Extract icao station
                }
                else if (datePat.IsMatches(part))
                {
                    // Date: ddhhmmZ
                    // DatePat "([0-9]{2})([0-9]{2})([0-9]{2})Z"  
                    int day = ToInt(datePat.groups[1]);
                    int hour = ToInt(datePat.groups[2]);
                    int min = ToInt(datePat.groups[3]);

                    reportTimeUTC = new DateTime(year, month, day, hour, min, 0);
                }
                else if (modPat.IsMatch(part))
                {
                    this.reportMod = part;      // Report Modifier: AUTO, COR, CCx or RRx
                }
                else if (windSpdPat.IsMatches(part))
                {
                    // windSpdPat ("([0-9]{3}|VRB)([0-9]{2,3})G?([0-9]{2,3})?(KT|MPS|KMH)");
                    if (windSpdPat.groups[1] == "VRB")
                        this.wind.Direction.Special = "Vrb";
                    else
                        this.wind.Direction.Degrees = ToDouble(windSpdPat.groups[1]);

                    double wndSpd = ToDouble(windSpdPat.groups[2]); 
                    this.wind.Speed.SetValue(wndSpd, windSpdPat.groups[4]);
                    if (wndSpd == 0)
                    {
                        this.wind.Direction.Special = ".";
                        this.wind.Speed.Special = "Calm";
                    }

                    if (windSpdPat.groups[3].Length > 0)
                    {
                        this.wind.Gust.SetValue(ToDouble(windSpdPat.groups[3]), windSpdPat.groups[4]); 
                    }
                }
                else if (windDirPat.IsMatches(part)) 
                {
                    // Variable wind-direction
                    this.wind.VarDirBeg.Degrees = ToDouble(windDirPat.groups[1]);
                    this.wind.VarDirEnd.Degrees = ToDouble(windDirPat.groups[2]);
                }
                else if (visPat.IsMatches(part))
                {
                    // visPat ("^([0-9]{4})(N|S|E|W|)$");
                    if (visibilityList.Count < MaxVisibilityGroup)
                    {
                        Visibility vis = new Visibility();

                        if (visPat.groups[1] == "0000")       // Visibility in meters (4 digits only)
                        {
                            // vis.prefix = -1; // Special case, Less than
                            vis.SetValue(50.0, (int)Visibility.Units.Meter);
                        }
                        else if (visPat.groups[1] == "9999")
                        {
                            // vis.perfix = 1;      // Special case, Greater than
                            vis.SetValue(10.0, (int)Visibility.Units.Kilometer);
                        }
                        else
                        {
                            double meters = ToDouble(datePat.groups[1]);
                            vis.SetValue(meters, (int)Visibility.Units.Meter);
                        }

                        if (visPat.groups.Length > 2 && visPat.groups[2].Length > 0)
                        {
                            // vis.SetDirection = visPat.group[2];
                        }

                        visibilityList.Add(vis);
                    }
                }
                else if (visDigPat.IsMatch(part))
                {
                    visMileStr = part.Trim() + " ";        // Visibility Group, single digit followed by space.
                }
                else if (visDisPat.IsMatches(visMileStr + part))
                {
                    // "^M?(([0-9]?)[ ]?([0-9])(/?)([0-9]*))SM$"
                    if (visibilityList.Count < MaxVisibilityGroup)
                    {
                        Visibility vis = new Visibility();

                        double miles = 0;
                        if (visDisPat.groups[4] == "/")
                        {
                            miles = (visDisPat.groups[2].Length > 0) ? ToDouble(visDisPat.groups[2]) : 0;
                            miles += ToDouble(visDisPat.groups[3]) / Math.Max(1.0, ToDouble(visDisPat.groups[5]));
                        }
                        else
                        {
                            miles = ToDouble(visDisPat.groups[1]);
                        }

                        if (visDisPat.groups[0][0] == 'M')
                        {
                            /* Prefix - less than */
                            vis.HowTo = Visibility.How.LessThan;
                        }
                       
                        vis.SetValue(miles, (int)Visibility.Units.Mile);
                        visibilityList.Add(vis);
                    }
                }
                else if (part == "CAVOK")
                {
                    if (visibilityList.Count < MaxVisibilityGroup)
                    {
                        /* CAVOK is used when the visibility is greater than 10
                        * kilometers, the lowest cloud-base is at 5000 feet or more
                        * and there is no significant weather.
                        */
                        Visibility vis = new Visibility();
                        vis.HowTo = Visibility.How.GreaterThan;
                        vis.SetValue(10, (int)Visibility.Units.Kilometer);
                        visibilityList.Add(vis);

#if false
                        this.clouds_array[0].Condition = "CAVOK";
#endif
                    }
                }
                else if (runwayPat.IsMatches(part))
                {
                    /* Runway-group */
                    // "^R([0-9]{2})([RLC]?)/([MP]?)([0-9]{4})([DNU]?)V?(P?)([0-9]{4})?([DNU]?)$"
                    if (runwayList.Count < MaxRunwayGroup)
                    {
                        Runway runway = new Runway();

                        runway.Id = runwayPat.groups[1];

                        if (runwayPat.groups[2] != "")
                        {
                            runway.Approach = runwayPat.groups[2];
                        }
                        if (runwayPat.groups[7] != "")
                        {
                            /* We have both min and max visibility since regs[7] holds
                            * the max visibility.
                            */
                            if (runwayPat.groups[5] != "")
                            {
                                /* regs[5] is tendency for min visibility. */
                                runway.MinTendency = runwayPat.groups[5];
                            }
                            if (runwayPat.groups[8] != "")
                            {
                                /* regs[8] is tendency for max visibility. */
                                runway.MaxTendency = runwayPat.groups[8];
                            }
                            if (runwayPat.groups[3] == "M")
                            {
                                /* Less than. */
                                runway.HowTo = Runway.How.LessThan;   //  = -1;
                            }
                            runway.MinLength.SetValue(ToDouble(runwayPat.groups[4]), (int)LDistance.Units.Meter);
                            if (runwayPat.groups[6] == "P")
                            {
                                /* Greater than. */
                                runway.HowTo = Runway.How.GreaterThan;   //  = 1;
                            }
                            runway.MaxLength.SetValue(ToDouble(runwayPat.groups[7]), (int)LDistance.Units.Meter);
                        }
                        else
                        {
                            /* We only have a single visibility. */
                            if (runwayPat.groups[5] != "")
                            {
                                /* regs[5] holds the tendency for visibility. */
                                runway.Tendency = runwayPat.groups[5];
                            }
                            if (runwayPat.groups[3] == "M")
                            {
                                /* Less than. */
                                runway.HowTo = Runway.How.LessThan;   //  = -1;
                            }
                            else if (runwayPat.groups[3] == "P")
                            {
                                /* Greater than. */
                                runway.HowTo = Runway.How.GreaterThan;   //  = 1;
                            }
                            runway.Length.SetValue(ToDouble(runwayPat.groups[4]), (int)LDistance.Units.Meter);
                        }
                        
                        this.runwayList.Add(runway);
                    }
                }
                else if (wxPat.IsMatches(part))
                {

                    // "^(VC)?" +       // Proximity
                    // "(-|\\+)?" +                         // Intensity
                    // "(MI|PR|BC|DR|BL|SH|TS|FZ)?" +       // Descriptor
                    // "((DZ|RA|SN|SG|IC|PL|GR|GS|UP)+)?" + // Precipitation
                    // "(BR|FG|FU|VA|DU|SA|HZ|PY)?" +       // Obscuration
                    // "(PO|SQ|FC|SS)?$",                   // Other

                    //    TS=THUN,
                    //    RA=RAIN,
                    //    DZ=RAIN,
                    //    SN=SNOW,
                    //    SG=SNOW,
                    //    GR=HAIL,
                    //    GS=HAIL,
                    //    PE=HAIL,
                    //    IC=HAIL,
                    //    BR=FOG,
                    //    FG=FOG


                    // Current weather-group.
                    if (weatherList.Count < MaxWeatherGroup)
                    {
                        Weather weather = new Weather();

                        weather.Proximity = wxPat.groups[1];
                        weather.Intensity = wxPat.groups[2];
                        weather.Descriptor = wxPat.groups[3];
                        weather.Precipitation = wxPat.groups[4];
                        weather.Obscuration = wxPat.groups[6];
                        weather.Other = wxPat.groups[7];

                        weatherList.Add(weather);
                    }
                }

                else if (part == "SKC" || part == "CLR")
                {
                    // Cloud-group
                    if (cloudsList.Count < MaxCloudGroup)
                    {
                        Clouds clouds = new Clouds();
                        clouds.Text = part;
                        cloudsList.Add(clouds);
                    }
                }
                else if (cloudPat.IsMatches(part))
                {
                    // "^(VV|FEW|SCT|BKN|OVC)([0-9]{3}|///)(CB|TCU)?$"
                    if (cloudsList.Count < MaxCloudGroup)
                    {
                        Clouds clouds = new Clouds();
                        clouds.Text = cloudPat.groups[1];
                        if (cloudPat.groups[3] != "")
                        {
                            clouds.Cumulus = cloudPat.groups[3];
                        }
                        if (cloudPat.groups[2] == "000")
                        {
                            /* '000' is a special height. */
                            clouds.Height.SetValue(100, (int)LDistance.Units.Feet);
                            clouds.HowTo = Clouds.How.LessThan;   //  = -1; 
                        }
                        else if (cloudPat.groups[2] == "///")
                        {
                            /* '///' means height nil */
                            clouds.Height.SetValue(double.NaN, 0);
                        }
                        else
                        {
                            clouds.Height.SetValue(ToDouble(cloudPat.groups[2]) * 100, (int)LDistance.Units.Feet);
                        }
                        cloudsList.Add(clouds);
                    }
                }
                else if (temDewPat.IsMatches(part))
                {
                    // Temperature/Dew Point Group.
                    // ^(M?[0-9]{2})/(M?[0-9]{2})?$"
                    string tStr = temDewPat.groups[1];
                    if (tStr.Length > 0)
                    {
                        double tCelcius = ToDouble(tStr.Replace("M", "-"));
                        temperature.SetValue(tCelcius, (int)Temperature.Units.Celcius);
                    }

                    string dStr = temDewPat.groups[2];
                    if (dStr.Length > 0)
                    {
                        double dewPt = ToDouble(dStr.Replace("M", "-"));
                        dewpoint.SetValue(dewPt, (int)Temperature.Units.Celcius); 
                    }
                }
                else if (pressApat.IsMatches(part))
                {
                    // Altimeter,  pressure measured in inHg.
                    this.pressure.SetValue(ToDouble(pressApat.groups[1]) / 100, (int)Pressure.Units.InchesHg);
                }
                else if (pressQpat.IsMatches(part))
                {
                    // "Q([0-9]{4})" 
                    // Altimeter - The specification doesn't say anything about  the Qxxxx-form, but it's in the METARs.
                    // The pressure measured in hPa  (hecoPascal = milliBar)

                    this.pressure.SetValue(ToDouble(pressApat.groups[1]) / 100, (int)Pressure.Units.MilliBar);
                }
 
            }

            return true;
        }


        /// <summary>
        ///  Decode METAR remarks section
        /// </summary>
        /// <param name="parts"></param>
        /// <returns></returns>
        private bool DecodeRemarks(string[] parts)
        {
            // Temperature and dewpoint in 10th celcius
            RegCompare tenthTandDpat = new RegCompare("^T([0-9]{4})([0-9]{4})");
            RegCompare tenthTpat = new RegCompare("^T([0-9]{4}$)");
            RegCompare precipPat = new RegCompare("^P([0-9]{4})");

            int partIdx = 0;
            foreach (string part in parts)
            {
                partIdx++;

                if (tenthTandDpat.IsMatches(part))
                {
                    // "^T([0-9]{4})([0-9]{4})"
                    // Temperature/Dew Point Group, coded to tenth of degree Celsius.
                    int num = ToInt(tenthTandDpat.groups[1]);
                    num = (num > 1000) ? -(num%1000) : num;
                    this.temperature.SetValue( num / 10.0, (int)Temperature.Units.Celcius);
                   
                    num = ToInt(tenthTandDpat.groups[2]);
                    num = (num > 1000) ? -(num % 1000) : num;
                    this.dewpoint.SetValue(num / 10.0, (int)Temperature.Units.Celcius);
                }
                else if (tenthTpat.IsMatches(part))
                {
                    // "^T([0-9]{4})"
                    // Temperature Group, coded to tenth of degree Celsius.
                    int num = ToInt(tenthTandDpat.groups[1]);
                    num = (num > 1000) ? -(num % 1000) : num;
                    this.temperature.SetValue(num / 10.0, (int)Temperature.Units.Celcius);
                }
                else if (precipPat.IsMatches(part))
                {
                    // "^P([0-9]{4})"
                    // Precipitation during last hour in hundredths of an inch

                    if (precipPat.groups[1] == "0000")
                    {
                        this.precipitation.SetValue(Precipitation.Trace, (int)Precipitation.Units.Inches);
                    }
                    else
                    {
                        this.precipitation.SetValue(ToDouble(precipPat.groups[1]) / 100.0,
                            (int)Precipitation.Units.Inches);
                    }
                }

#if false
                else if (ClassEreg.ereg("^1([0-9]{4}$)", part, ref regs))
                {
                    /*
                    * 6 hour maximum temperature Celsius, coded to tenth of degree
                    */
                    this.tempMinMax6h.Tenth = ((int)(System.Convert.ToDouble(regs[1]) / 10)).ToString();
                }
                else if (ClassEreg.ereg("^2([0-9]{4}$)", part, ref regs))
                {
                    /*
                    * 6 hour minimum temperature Celsius, coded to tenth of degree
                    */
                    this.tempMinMax6h.Tenth = ((int)(System.Convert.ToDouble(regs[1]) / 10)).ToString();
                }
                else if (ClassEreg.ereg("^4([0-9]{4})([0-9]{4})$", part, ref regs))
                {
                    /*
                    * 24 hour maximum and minimum temperature Celsius, coded to
                    * tenth of degree
                    */
                    this.tempMinMax24h.Tenth = ((int)(System.Convert.ToDouble(regs[1]) / 10)).ToString();
                    this.tempMinMax24h.Tenth = ((int)(System.Convert.ToDouble(regs[2]) / 10)).ToString();
                }

                else if (ClassEreg.ereg("^6([0-9]{4})", part, ref regs))
                {
                    /*
                    * Precipitation during last 3 or 6 hours in hundredths of an
                    * inch.
                    */
                    if (regs[1] == "0000")
                    {
                        this.precipitation.height_6h.Inch = -1;
                    }
                    else
                    {
                        this.precipitation.height_6h.Inch = System.Convert.ToDouble(regs[1]) / 100;
                    }
                }
                else if (ClassEreg.ereg("^7([0-9]{4})", part, ref regs))
                {
                    /*
                    * Precipitation during last 24 hours in hundredths of an inch.
                    */
                    if (regs[1] == "0000")
                    {
                        this.precipitation.height_24h.Inch = -1;
                    }
                    else
                    {
                        this.precipitation.height_24h.Inch = System.Convert.ToDouble(regs[1]) / 100;
                    }
                }
                else if (ClassEreg.ereg("^4/([0-9]{3})", part, ref regs))
                {
                    /*
                    * Snow depth in inches
                    */
                    if (regs[1] == "0000")
                    {
                        this.precipitation.height_snow.Inch = -1;
                    }
                    else
                    {
                        this.precipitation.height_snow.Inch = System.Convert.ToDouble(regs[1]);
                    }
                }
                else
                {
                    /*
                    * If we couldn't match the group, we assume that it was a
                    * remark.
                    */
                    this.remarks += " " + part;
                }
#endif

            }

            return true;
        }

        private void DerivedParameters()
        {

#if false
            /*
            * Relative humidity
            */
            if ((this.temperature.current.Celcius != 0) && (this.temperature.dew.Celcius != 0))
            {
                this.humidity.b_no_data = false;
                this.humidity.relative = System.Math.Pow(10, (1779.75 * (this.temperature.dew.Celcius -
                    this.temperature.current.Celcius)
                    / ((237.3 + this.temperature.dew.Celcius) *
                    (237.3 + this.temperature.current.Celcius))
                    + 2));
            }
            /*
            *  Compute windchill if temp < 40f and windspeed > 3 mph
            */
            if ((this.temperature.current.Fahrenheit != 0) &&
                this.temperature.current.Fahrenheit <= 40 &&
                (this.wind.speed.Mileshours != 0) &&
                this.wind.speed.Mileshours > 3
                )
            {
                this.temperature.windchill.Fahrenheit = 35.74 + 0.6215 * this.temperature.current.Fahrenheit
                    - 35.75 * System.Math.Pow(this.wind.speed.Mileshours, 0.16)
                    + 0.4275 * this.temperature.current.Fahrenheit *
                    System.Math.Pow(this.wind.speed.Mileshours, 0.16);
            }
            /*
            * Compute heat index if temp > 70F
            */
            if ((this.temperature.current.Fahrenheit != 0) &&
                this.temperature.current.Fahrenheit > 70 &&
                (this.humidity.relative != 0)
                )
            {
                this.temperature.heatindex.Fahrenheit = -42.379
                    + 2.04901523 * this.temperature.current.Fahrenheit
                    + 10.1433312 * this.humidity.relative
                    - 0.22475541 * this.temperature.current.Fahrenheit
                    * this.humidity.relative
                    - 0.00683783 * this.temperature.current.Fahrenheit
                    * this.temperature.current.Fahrenheit
                    - 0.05481717 * this.humidity.relative
                    * this.humidity.relative
                    + 0.00122874 * this.temperature.current.Fahrenheit
                    * this.temperature.current.Fahrenheit
                    * this.humidity.relative
                    + 0.00085282 * this.temperature.current.Fahrenheit
                    * this.humidity.relative
                    * this.humidity.relative
                    - 0.00000199 * this.temperature.current.Fahrenheit
                    * this.temperature.current.Fahrenheit
                    * this.humidity.relative
                    * this.humidity.relative;
            }
            /*
            * Compute the humidity index
            */
            if (this.humidity.relative != 0)
            {
                double e = (6.112 * System.Math.Pow(10, 7.5 * this.temperature.current.Celcius
                    / (237.7 + this.temperature.current.Celcius))
                    * this.humidity.relative / 100) - 10;
                this.humidity.index.Celcius = this.temperature.current.Celcius + 5.0 / 9 * e;
            }
#endif
        }

    }
}
