﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

// printing
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Drawing.Printing;
using System.Drawing.Drawing2D;
using System.Printing;
using System.IO;
using System.Windows.Forms;
using System.Drawing.Imaging;


namespace WebMetar_NS
{
    /// <summary>
    /// Helper class to print our User components and standard UI objects.
    /// 
    /// Author: Dennis Lang 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// </summary>
    public class PrintIt
    {
        // Print fields
        private PrintDocument m_printDoc = new PrintDocument();
        private PageSetupDialog m_setupDlg = new PageSetupDialog();
        private PrintPreviewDialog m_previewDlg = new PrintPreviewDialog();
        private PrintDialog m_printDlg = new PrintDialog();

        private int m_nPageNumber = 1;
        private UserControl m_userControl;
        private string m_title;
        private Rectangle m_screenRect;
        private Bitmap m_bitmap;

        public PrintIt()
        {
            m_printDoc.DocumentName = "WebMetar";
            m_setupDlg.Document = m_printDoc;
            m_previewDlg.Document = m_printDoc;
            m_printDlg.Document = m_printDoc;

            m_setupDlg.PageSettings.Landscape = false;
            m_printDoc.BeginPrint += new PrintEventHandler(OnBeginPrint);
            m_printDoc.PrintPage += new PrintPageEventHandler(OnPrintPage);
        }

        /// <summary>
        ///	Show the standard page setup dialog box that lets the user specify
        ///	margins, page orientation, page sources, and paper sizes.
        /// </summary>
        public DialogResult PageSetup()
        {
            return m_setupDlg.ShowDialog();
        }

        /// <summary>
        ///		Show the standard print preview dialog box.
        /// </summary>
        public DialogResult PrintPreview(IWin32Window owner, UserControl userControl, string title)
        {
            m_userControl = userControl;
            m_title = title;
            return PrintPreview();
        }
        /// <summary>
        ///		Show the standard print preview dialog box.
        /// </summary>
        public DialogResult PrintPreview(IWin32Window owner, Rectangle captureRect, string title)
        {
            m_userControl = null;
            m_screenRect = captureRect;
            GrabScreenImage(m_screenRect);
            m_title = title;
            return PrintPreview();
        }

        private DialogResult PrintPreview()
        {
            m_nPageNumber = 1;
            m_previewDlg.Width = m_setupDlg.PageSettings.Landscape ? 1000 : 600;
            m_previewDlg.Height = 600;

            return m_previewDlg.ShowDialog();
        }

        /// <summary>
        /// Start the print process.
        /// </summary>
        public void Print(IWin32Window owner, UserControl userControl, string title)
        {
            m_userControl = userControl;
            m_title = title;
            Print();
        }

        /// <summary>
        ///	 Start the print process.
        /// </summary>
        public void Print(IWin32Window owner, Rectangle captureRect, string title)
        {
            m_userControl = null;
            m_screenRect = captureRect;
            GrabScreenImage(m_screenRect);
            m_title = title;
            Print();
        }

        private void Print()
        {
            m_nPageNumber = 1;
            m_printDlg.UseEXDialog = true;
            m_printDlg.AllowSomePages = true;
            m_printDlg.AllowCurrentPage = true;
            m_printDlg.AllowPrintToFile = true;
            m_printDlg.ShowHelp = true;
            m_printDlg.ShowNetwork = true;
            m_printDlg.AllowSelection = false; //  m_listView.SelectedItems.Count > 0;

            // Show the standard print dialog box, that lets the user select a printer
            // and change the settings for that printer.
            if (m_printDlg.ShowDialog() == DialogResult.OK)
            {
                m_printDoc.Print();
            }
        }

        /// <summary>
        /// Capture image of dialog (screen shot)
        /// </summary>
        /// <returns></returns>
        public void GrabScreenImage(Rectangle rect)
        {
            int color = Screen.PrimaryScreen.BitsPerPixel;
            PixelFormat pFormat;

            switch (color)
            {
                case 8:
                case 16:
                    pFormat = PixelFormat.Format16bppRgb565;
                    break;

                case 24:
                    pFormat = PixelFormat.Format24bppRgb;
                    break;

                case 32:
                    pFormat = PixelFormat.Format32bppArgb;
                    break;

                default:
                    pFormat = PixelFormat.Format32bppArgb;
                    break;
            }

            Application.DoEvents();
            m_bitmap = new Bitmap(rect.Width, rect.Height, pFormat);
            Graphics g = Graphics.FromImage(m_bitmap);
            g.CopyFromScreen(rect.Left, rect.Top, 0, 0, rect.Size);
            g.Dispose();
        }


        #region Events Handlers
        private void OnBeginPrint(object sender, PrintEventArgs e)
        {
            PreparePrint();
        }

        private void OnPrintPage(object sender, PrintPageEventArgs e)
        {
            Graphics g = e.Graphics;

            // Printable area (approximately) of the page, taking into account the user margins
            RectangleF rectFull;			  // The full available space
            rectFull = new RectangleF(
                e.MarginBounds.Left - (e.PageBounds.Width - g.VisibleClipBounds.Width) / 2,
                e.MarginBounds.Top - (e.PageBounds.Height - g.VisibleClipBounds.Height) / 2,
                e.MarginBounds.Width,
                e.MarginBounds.Height);

            Font hdrFont = new Font(FontFamily.GenericSerif, 12);   // new Font(Font.m_userControl, FontStyle.Bold);
            RectangleF rectBody = RectangleF.Inflate(rectFull, 0, -8 * hdrFont.GetHeight(g));

            if (m_userControl == null)
            {
                double xScale = (double)m_bitmap.Width / rectBody.Width;
                double yScale = (double)m_bitmap.Height / rectBody.Height;
                if (xScale < yScale)
                    rectBody.Width = (int)(xScale * rectBody.Width / yScale);
                else
                    rectBody.Height = (int)(yScale * rectBody.Height / xScale);
                e.Graphics.PageUnit = GraphicsUnit.Display;
                e.Graphics.DrawImage(m_bitmap, rectBody);
            }
            else if (m_userControl is BarGraph)
            {
                BarGraph barGraph = (BarGraph)m_userControl;
                barGraph.PrintToGraphics(e.Graphics, rectBody /* e.MarginBounds */);
            }
            else if (m_userControl is TimeGraph)
            {
                TimeGraph barGraph = (TimeGraph)m_userControl;
                barGraph.PrintToGraphics(e.Graphics, rectBody /* e.MarginBounds */);
            }
            else
            {
                Bitmap bitmap = new Bitmap(m_userControl.Width, m_userControl.Height);
                m_userControl.DrawToBitmap(bitmap, new Rectangle(0, 0, bitmap.Width, bitmap.Height));
                double xScale = (double)bitmap.Width / rectBody.Width;
                double yScale = (double)bitmap.Height / rectBody.Height;
                if (xScale < yScale)
                    rectBody.Width = (int)(xScale * rectBody.Width / yScale);
                else
                    rectBody.Height = (int)(yScale * rectBody.Height / xScale);
                e.Graphics.PageUnit = GraphicsUnit.Display;
                e.Graphics.DrawImage(bitmap, rectBody);
            }
         
            // Display title at top
            StringFormat sfmt = new StringFormat();
            sfmt.Alignment = StringAlignment.Center;

            // Font hdrFont = new Font(m_listView.Font.FontFamily, m_listView.Font.SizeInPoints * 1.5f);
            g.DrawString(m_title, hdrFont, Brushes.Black, rectFull, sfmt);

            // Display page number at bottom
            sfmt.LineAlignment = StringAlignment.Far;
            string pageStr = string.Format("{0}    Page {1}",
                DateTime.Now.ToShortDateString(), m_nPageNumber);
            g.DrawString(pageStr,hdrFont, Brushes.Black, rectFull, sfmt);


            e.HasMorePages = false;
            m_nPageNumber = 1;
        }
        #endregion

        private void PreparePrint()
        {
            // Graphics g = m_userControl.CreateGraphics();
        }
    }
}
