﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WebMetar_NS
{
    ///
    /// Extract fields from METAR text and store in METAR class.
    /// 
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// Assumes text format found at ftp site (available Jan 2010).
    /// 
    ///   "ftp://tgftp.nws.noaa.gov/data/observations/metar/decoded/{0}.TXT";
    ///   
    /// Example response:
    /// 
    ///    BOSTON LOGAN INTERNATIONAL, MA, United States (KBOS) 42-22N 071-01W 54M
    ///    Jan 13, 2010 - 08:54 PM EST / 2010.01.14 0154 UTC
    ///    Wind: from the WNW (300 degrees) at 13 MPH (11 KT):0
    ///    Visibility: 10 mile(s):0
    ///    Visibility: 2 1/2 mile(s):0
    ///    Sky conditions: overcast
    ///    Temperature: 27.0 F (-2.8 C)
    ///    Windchill: 15 F (-9 C):1
    ///    Dew Point: 15.1 F (-9.4 C)
    ///    Relative Humidity: 60%
    ///    Pressure (altimeter): 30.12 in. Hg (1019 hPa)
    ///    Precipitation last hour: 0.04 inches
    ///    ob: KBOS 140154Z 30011KT 10SM SCT065 BKN080 OVC110 M03/M09 A3012 RMK AO2 SLP197 T10281094
    ///    cycle: 2
    ///    
    class TextToMetar
    {
        public static bool Decode(string text, ref Metar metar)
        {
            RegCompare icaoPat = new RegCompare(@".* \(([A-Z][A-Z][A-Z0-9][A-Z0-9])\)");
            RegCompare timePat = new RegCompare(@"[^/]*/ +(20[0-9][0-9])\.([0-9][0-9])\.([0-9][0-9]) +([0-9]{4}) UTC");
            RegCompare windPat = new RegCompare(@"Wind[: ]+(calm|from the ([NEWS]+)  +\(([0-9]+) degrees\) at ([0-9]+) MPH)");
            RegCompare vis1Pat = new RegCompare(@"Visibility[: ]+([0-9]+) mile");
            RegCompare vis2Pat = new RegCompare(@"Visibility[: ]+([0-9]*)( [0-9]+)/([0-9]+) mile");
            RegCompare skyPat = new RegCompare(@"Sky conditions[: ]+([a-zA-Z0-9]+)");
            RegCompare temPat = new RegCompare(@"Temperature[: ]+([0-9.-]+) F");
            RegCompare dewPat = new RegCompare(@"Dew Point[: ]+([0-9.-]+) F");
            RegCompare humPat = new RegCompare(@"Relative Humidity[: ]+([0-9]+)\%");
            RegCompare pressPat = new RegCompare(@"Pressure .*[: ]+([0-9.-]+) in\. Hg");
            RegCompare precipPat = new RegCompare(@"Precipitation .*[: ]+([0-9.-]+) inches");
            RegCompare obPat = new RegCompare(@"ob: (.*)");

            int fieldCnt = 0;
            string[] rows = text.Split('\n');
            foreach (string row in rows)
            {
                if (icaoPat.IsMatches(row))
                {
                    metar.icao = icaoPat.groups[1];
                    fieldCnt += 100;
                }
                else if (timePat.IsMatches(row))
                {
                    int year = ToInt(timePat.groups[1]);
                    int mon = ToInt(timePat.groups[2]);
                    int day = ToInt(timePat.groups[3]);
                    int hhmm = ToInt(timePat.groups[4]);

                    DateTime dt = new DateTime(year, mon, day, hhmm / 100, hhmm % 100, 0);
                    metar.reportTimeUTC = dt;
                    fieldCnt += 100;
                }
                else if (windPat.IsMatches(row))
                {
                    if (windPat.groups.Length < 4)
                    {
                        metar.wind.Speed.SetValue(0, (int)Speed.Units.MilesPerHour);
                    }
                    else
                    {
                        metar.wind.Direction.Degrees = ToInt(windPat.groups[3]);
                        metar.wind.Speed.SetValue(ToInt(windPat.groups[4]), (int)Speed.Units.MilesPerHour);
                    }
                    fieldCnt++;
                }
                else if (vis1Pat.IsMatches(row))
                {
                    // RegCompare visPat = new RegCompare(@"Visibility: ([0-9]+) mile");
                    Visibility vis = new Visibility();
                    vis.SetValue(ToDouble(vis1Pat.groups[1]), (int)LDistance.Units.Mile);
                    metar.visibilityList.Add(vis);
                    fieldCnt++;
                }
                else if (vis2Pat.IsMatches(row))
                {
                    // "Visibility: ([0-9]*)( [0-9]+)/([0-9]+) mile");
                    Visibility vis = new Visibility();
                    double miles = 0;
                    if (vis2Pat.groups[1].Length > 0)
                        miles = ToDouble(vis2Pat.groups[1]);
                    miles += ToDouble(vis2Pat.groups[2]) / Math.Max(1.0, ToDouble(vis2Pat.groups[3]));
                    vis.SetValue(miles, (int)LDistance.Units.Mile);
                    metar.visibilityList.Add(vis);
                    fieldCnt++;
                }
                else if (skyPat.IsMatches(row))
                {
                    // RegCompare skyPat = new RegCompare(@"Sky conditions: ([a-zA-Z0-9]+)");
                    Clouds clouds = new Clouds();
                    clouds.Text = skyPat.groups[1];
                    metar.cloudsList.Add(clouds);
                    fieldCnt++;
                }
                else if (temPat.IsMatches(row))
                {
                    // RegCompare temPat = new RegCompare(@"Temperature: ([0-9.-]+ F");
                    metar.temperature.SetValue(ToDouble(temPat.groups[1]), (int)Temperature.Units.Farhrenheit);
                    fieldCnt += 10;
                }
                else if (dewPat.IsMatches(row))
                {
                    // RegCompare dewPat = new RegCompare(@"Dew Point: ([0-9.-]+ F");
                    metar.dewpoint.SetValue(ToDouble(dewPat.groups[1]), (int)Temperature.Units.Farhrenheit);
                    fieldCnt++;
                }
                else if (humPat.IsMatches(row))
                {
                    // RegCompare humPat = new RegCompare(@"Relative Humidify: ([0-9]+)\%");
                    metar.humidity.SetValue(ToDouble(humPat.groups[1]), (int)Humidity.Units.Percent);
                    fieldCnt++;
                }
                else if (pressPat.IsMatches(row))
                {
                    // RegCompare pressPat = new RegCompare(@"Pressure .*: ([0-9.-]+ in\. Hg");
                    metar.pressure.SetValue(ToDouble(pressPat.groups[1]), (int)Pressure.Units.InchesHg);
                    fieldCnt += 10;
                }
                else if (precipPat.IsMatches(row))
                {
                    // RegCompare pressPat = new RegCompare(@"Precipitation .*: ([0-9.-]+ inches");
                    metar.precipitation.SetValue(ToDouble(precipPat.groups[1]), (int)Precipitation.Units.Inches);
                    fieldCnt += 10;
                }
                else if (obPat.IsMatches(row))
                {
                    // RegCompare obPat = new RegCompare(@"ob: (.*)");
                    metar.metarStr = obPat.groups[1] + " (text)";
                    fieldCnt += 10;
                }
            }

            // Field count indicates how well we decode text report.
            // Field count should be 230 or greater for success.
            //
            //   icao + time = 200
            //   temp + press + obs = 30
            //   everything else is +1

            return (fieldCnt >= 230);
        }

        static double ToDouble(string s)
        {
            try
            {
                return double.Parse(s);
            }
            catch
            {
                return double.NaN;
            }
        }
        static int ToInt(string s)
        {
            try
            {
                return int.Parse(s);
            }
            catch
            {
                return 0;
            }
        }
    }
}
