﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WebMetar_NS
{
    /// 
    /// WxBase is base class for Weather parameters.
    /// Supports:
    ///     * Unit enum and string array list (of valid choices)
    ///     * Fromat value and string array list (of valid choices)
    ///     * Unit conversion (via array of conversion delegates)
    ///     * Text presentation with and without units.
    ///     
    /// Author: Dennis Lang  2010      
    /// http://home.comcast.net/~lang.dennis/
    /// 
    #region ==== WxBase
    public delegate double Conversion(double val, int toUnit);

    [Serializable]
    public class WxBase
    {
        public WxBase(string[] unitStrs, Conversion[] cnvDelegates)
        {
            m_unitStrs = unitStrs;
            m_cnvDelegates = cnvDelegates;
            m_unit = 0;
            m_value = double.NaN;
        }

        protected string[] m_unitStrs;
        protected Conversion[] m_cnvDelegates;

        protected double m_value;
        protected int m_unit;
        protected string m_noDataStr = string.Empty;
        protected string m_fmtStr = "F0"; // "F1", "F2", ...
        protected string[] m_fmtArray = new string[] { "F0", "F1", "F2" };
        protected Dictionary<double, string> m_lookup = new Dictionary<double, string>();

        public void SetValue(double val, int unit)
        {
            m_value = val;
            m_unit = unit;
        }

        public void SetValue(double val, string unitStr)
        {
            m_value = val;
            // m_unit = Array.FindIndex<string>(m_unitStrs, delegate(string s) {return s == unitStr; });
            // m_unit = Array.FindIndex<string>(m_unitStrs, x => x[0].Equals(unitStr));
            m_unit = Array.FindIndex<string>(m_unitStrs, x => string.Equals(x, unitStr, StringComparison.OrdinalIgnoreCase));
            if (m_unit == -1)
                throw new Exception("unknown unit " + unitStr);
        }

        public void SetLookup(double val, string text)
        {
            m_lookup[val] = text;
        }

        public string NoData
        {
            get { return m_noDataStr; }
            set { m_noDataStr = value; }
        }
        public string Format
        {
            get { return m_fmtStr; }
            set { m_fmtStr = value; }
        }

        public string[] FormatArray
        {
            get { return m_fmtArray; }
            set { m_fmtArray = value; }
        }
        public double GetValue(int unit)
        {
            return m_cnvDelegates[m_unit](m_value, unit);
        }

        public bool IsPopulated 
        {
            get { return (double.IsNaN(m_value) == false); }
        }
        public virtual bool IsNumber
        {
            get { return true; }
        }

        public int Unit
        {
            get { return m_unit; }
            set { m_unit = value; }
        }

        public string[] UnitArray
        {
            get { return m_unitStrs; }
        }

        public string UnitText
        {
            get { return m_unitStrs[m_unit]; }
        }

        public virtual double Value
        {
            get { return m_value; }
            set { m_value = value; }
        }

        public virtual string Text
        {
            get { return (IsPopulated ? Value.ToString(m_fmtStr) : m_noDataStr );  }
            set { try { m_value = double.Parse(value); } catch { m_value = double.NaN; } }
        }

        public virtual string TextandUnit
        {
            get { 
                    if (IsPopulated)
                    {
                        return m_lookup.ContainsKey(m_value) ?
                            m_lookup[m_value] : Text + UnitText;
                    }
                    return m_noDataStr;
                }
        }

       
        public WxBase ConvertTo(int unit, string fmt)
        {
            WxBase wxdata = this;
            wxdata.SetValue(m_cnvDelegates[m_unit](m_value, unit), unit);
            if (fmt != null)
                wxdata.Format = fmt;
            return wxdata;
        }
    }
    #endregion

}
