﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace WebMetar_NS
{
    /// 
    /// Class for numeric Weather parameters.
    /// Supports:
    ///     * Unit enum and string array list (of valid choices)
    ///     * Fromat value and string array list (of valid choices)
    ///     * Unit conversion (via array of conversion delegates)
    ///     * Text presentation with and without units.
    ///     
    /// Author: Dennis Lang  2009       
    /// http://home.comcast.net/~lang.dennis/
    /// 

#if false
    class WxNumber : WxBase
    {
        // Not currently part of implementation.
        // TODO - move more code into WxNumber and derive specific wx parameters from it.
    }
#endif

    #region ==== Base Numeric classes
    [Serializable]
    class Temperature : WxBase
    {
        public enum Units { Farhrenheit = 0, Celcius };
        static string[] unitStr = new string[] { "°F", "°C" };
        static private double FtoUnit(double f, int toUnit)
        {
            if (toUnit == (int)Units.Farhrenheit)
                return f;

            return (f - 32) * (5 / 9.0);
        }
        static private double CtoUnit(double c, int toUnit)
        {
            if (toUnit == (int)Units.Celcius)
                return c;
            return c * (9.0 / 5) + 32;
        }
        static Conversion[] cnvDelegates = new Conversion[] { FtoUnit, CtoUnit };

        public Temperature()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    [Serializable]
    class Speed : WxBase
    {
        public enum Units { MetersPerSecond = 0, KmPerHour, FeetPerSecond, MilesPerHour, Knots };
        static string[] unitStr = new string[] { "m/s", "Km/h", "Ft/s", "Mph", "Kt" };

        // Meters Per Second
        static public double MpstoUnit(double mps, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.MetersPerSecond:
                    return mps;
                case (int)Units.KmPerHour:
                    return mps * 3600 / 1000.0;
                case (int)Units.FeetPerSecond:
                    return LDistance.MtoUnit(mps, (int)LDistance.Units.Feet);
                case (int)Units.MilesPerHour:
                    return LDistance.MtoUnit(mps * 3600, (int)LDistance.Units.Mile);
                case (int)Units.Knots:
                    return LDistance.MtoUnit(mps * 3600, (int)LDistance.Units.NauticalMiles);
            }

            throw (new Exception("unknown units"));
        }

        // Kilometers per hour
        static public double KmphtoUnit(double Kmph, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.MetersPerSecond:
                    return Kmph * 1000 / 3600.0;
                case (int)Units.KmPerHour:
                    return Kmph;
                case (int)Units.FeetPerSecond:
                    return LDistance.KmtoUnit(Kmph / 3600.0, (int)LDistance.Units.Feet);
                case (int)Units.MilesPerHour:
                    return LDistance.KmtoUnit(Kmph, (int)LDistance.Units.Mile);
                case (int)Units.Knots:
                    return LDistance.KmtoUnit(Kmph, (int)LDistance.Units.NauticalMiles);
            }

            throw (new Exception("unknown units"));
        }

        // Feet per second
        static private double ftpstoUnit(double ftps, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.MetersPerSecond:
                    return LDistance.FttoUnit(ftps, (int)LDistance.Units.Meter);
                case (int)Units.KmPerHour:
                    return LDistance.FttoUnit(ftps, (int)LDistance.Units.Kilometer);
                case (int)Units.FeetPerSecond:
                    return ftps;
                case (int)Units.MilesPerHour:
                    return LDistance.FttoUnit(ftps * 3600, (int)LDistance.Units.Mile);
                case (int)Units.Knots:
                    return LDistance.FttoUnit(ftps * 3600, (int)LDistance.Units.NauticalMiles);
            }

            throw (new Exception("unknown units"));
        }

        // Miles per second
        static private double MphtoUnit(double Mph, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.MetersPerSecond:
                    return LDistance.MitoUnit(Mph / 3600.0, (int)LDistance.Units.Meter);
                case (int)Units.KmPerHour:
                    return LDistance.MitoUnit(Mph, (int)LDistance.Units.Kilometer);
                case (int)Units.FeetPerSecond:
                    return LDistance.MitoUnit(Mph / 3600.0, (int)LDistance.Units.Feet);
                case (int)Units.MilesPerHour:
                    return Mph;
                case (int)Units.Knots:
                    return LDistance.FttoUnit(Mph, (int)LDistance.Units.NauticalMiles);
            }

            throw (new Exception("unknown units"));
        }

        // Knots (nautical miles per hour)
        static private double KttoUnit(double kt, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.MetersPerSecond:
                    return LDistance.NMtoUnit(kt / 3600.0, (int)LDistance.Units.Meter);
                case (int)Units.KmPerHour:
                    return LDistance.NMtoUnit(kt, (int)LDistance.Units.Kilometer);
                case (int)Units.FeetPerSecond:
                    return LDistance.NMtoUnit(kt / 3600.0, (int)LDistance.Units.Feet);
                case (int)Units.MilesPerHour:
                    return LDistance.NMtoUnit(kt, (int)LDistance.Units.Mile);
                case (int)Units.Knots:
                    return kt;
            }

            throw (new Exception("unknown units"));
        }

        static Conversion[] cnvDelegates = new Conversion[] { MpstoUnit, KmphtoUnit, ftpstoUnit, MphtoUnit, KttoUnit };

        public Speed()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    /// <summary>
    /// Long Distances
    /// </summary>
    [Serializable]
    class LDistance : WxBase
    {
        public enum Units { Meter = 0, Kilometer, Feet, Mile, NauticalMiles }
        static string[] unitStr = new string[] { "m", "Km", "ft", "Mi", "Nm" };

        const double KmPerMeter = 1e-3;
        const double FeetPerMeter = 3.2808;
        const double MilesPerMeter = 6.2e-4;
        const double FeetPerMile = 5280;
        const double MilesPerKm = 0.621;
        const double MeterPerFeet = 0.304;
        const double MilesPerNautical = 1.15077945;
        const double KmPerNautical = 1.852;


        static public double FttoUnit(double ft, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Meter:
                    return ft * MeterPerFeet;
                case (int)Units.Kilometer:
                    return ft * MeterPerFeet * 1e-3;
                case (int)Units.Feet:
                    return ft;
                case (int)Units.Mile:
                    return ft / FeetPerMile;
                case (int)Units.NauticalMiles:
                    return ft / FeetPerMile / MilesPerNautical;
            }

            throw (new Exception("unknown units"));
        }

        static public double MitoUnit(double mi, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Meter:
                    return mi / MilesPerMeter;
                case (int)Units.Kilometer:
                    return mi / MilesPerMeter * 1e-3;
                case (int)Units.Feet:
                    return mi * FeetPerMile;
                case (int)Units.Mile:
                    return mi;
                case (int)Units.NauticalMiles:
                    return mi / MilesPerNautical;

            }

            throw (new Exception("unknown units"));
        }

        static public double MtoUnit(double m, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Meter:
                    return m;
                case (int)Units.Kilometer:
                    return m * 1e-3; ;
                case (int)Units.Feet:
                    return m * FeetPerMeter;
                case (int)Units.Mile:
                    return m * 1e-3 * MilesPerKm;
                case (int)Units.NauticalMiles:
                    return m * KmPerMeter / KmPerNautical;
            }

            throw (new Exception("unknown units"));
        }

        static public double KmtoUnit(double km, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Meter:
                    return km * 1000;
                case (int)Units.Kilometer:
                    return km;
                case (int)Units.Feet:
                    return km * 1e3 * FeetPerMeter;
                case (int)Units.Mile:
                    return km * MilesPerKm;
                case (int)Units.NauticalMiles:
                    return km / KmPerNautical;
            }

            throw (new Exception("unknown units"));
        }

        static public double NMtoUnit(double nm, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Meter:
                    return nm;              // TODO
                case (int)Units.Kilometer:
                    return nm;              // TODO
                case (int)Units.Feet:
                    return nm;              // TODO
                case (int)Units.Mile:
                    return nm;              // TODO
                case (int)Units.NauticalMiles:
                    return nm;
            }

            throw (new Exception("unknown units"));
        }
        static Conversion[] cnvDelegates = new Conversion[] { MtoUnit, KmtoUnit, FttoUnit, MitoUnit, NMtoUnit };

        public LDistance()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    [Serializable]
    class Percent : WxBase
    {
        public enum Units { Percent };
        static string[] unitStr = new string[] { "%" };

        static private double PerToUnit(double per, int toUnit)
        {
            return per;
        }
        static Conversion[] cnvDelegates = new Conversion[] { PerToUnit };

        public Percent()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    [Serializable]
    class Pressure : WxBase
    {
        public enum Units { InchesHg = 0, MilliBar, Atmosphere };
        static string[] unitStr = new string[] { "In", "mB", "Atm" };

        const double mBPerInHg = 33.863886;
        const double atmPerInHg = 0.033421;
        const double mBPerAtm = 1013.250;

        static private double IntoUnit(double inches, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.InchesHg:
                    return inches;              
                case (int)Units.MilliBar:
                    return inches * mBPerInHg;
                case (int)Units.Atmosphere:
                    return inches * atmPerInHg;              
            }

            throw (new Exception("unknown units"));
        }

        static private double mBtoUnit(double mB, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.InchesHg:
                    return mB / mBPerInHg;
                case (int)Units.MilliBar:
                    return mB;
                case (int)Units.Atmosphere:
                    return mB / mBPerAtm;
            }

            throw (new Exception("unknown units"));
        }

        static private double atmtoUnit(double atm, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.InchesHg:
                    return atm / atmPerInHg;
                case (int)Units.MilliBar:
                    return atm * mBPerAtm;
                case (int)Units.Atmosphere:
                    return atm;
            }

            throw (new Exception("unknown units"));
        }

        static Conversion[] cnvDelegates = new Conversion[] { IntoUnit, mBtoUnit, atmtoUnit };

        public Pressure()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    /// <summary>
    ///  Small Distances
    /// </summary>
    [Serializable]
    class SDistance : WxBase
    {
        public enum Units { Inches = 0, Millimeters, Centimeters };
        static string[] unitStr = new string[] { "in", "mm", "cm" };

        static private double IntoUnit(double inches, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Inches:
                    return inches;
                case (int)Units.Centimeters:
                    return inches * 2.54;
                case (int)Units.Millimeters:
                    return inches * 0.254;
            }
            throw new Exception("unknown unit " + toUnit.ToString());

        }
        static private double MmtoUnit(double mm, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Inches:
                    return mm / 0.254;
                case (int)Units.Centimeters:
                    return mm / 10.0;
                case (int)Units.Millimeters:
                    return mm; ;
            }
            throw new Exception("unknown unit " + toUnit.ToString());
        }
        static private double CmtoUnit(double cm, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Inches:
                    return cm / 2.54;
                case (int)Units.Centimeters:
                    return cm;
                case (int)Units.Millimeters:
                    return cm * 10;
            }
            throw new Exception("unknown unit " + toUnit.ToString());
        }
        static Conversion[] cnvDelegates = new Conversion[] { IntoUnit, MmtoUnit, CmtoUnit };

        public SDistance()
            : base(unitStr, cnvDelegates)
        {
        }
    }

    [Serializable]
    class Angle : WxBase
    {
        public enum Units { Degrees };
        static public double DtoUnit(double deg, int unit)
        {
            return deg;
        }

        static string[] unitStr = new string[] { "°" };
        static string[] fmtArray = new string[] { "F0", "F1", "F2"};
        static Conversion[] cnvDelegates = new Conversion[] { DtoUnit };

        public Angle()
            : base(unitStr, cnvDelegates)
        {
            m_value = double.NaN;
            m_fmtArray = fmtArray;
        }
        public Angle(string fmtStr)
            : base(unitStr, cnvDelegates)
        {
            m_fmtStr = fmtStr;
            m_value = double.NaN;
            m_fmtArray = fmtArray;
        }

        public double Degrees
        {
            get { return m_value; }
            set { m_value = value; }
        }
    }

    [Serializable]
    class Direction : WxBase
    {
        public enum Units { Degrees };
        static public double DtoUnit(double deg, int unit)
        {
            return deg;
        }

        //      NNW     N  NNE    NE
        //  NW          0  .    45
        //              |     /
        // WNW          | . /   . ENE
        //              | / . 
        // W 270 -------+-------  90 E
        //              |
        //  WSW         |         ESE
        //              |        SE
        //    SW   SSW 180 SSE
        //              S
        static string[] unitStr = new string[] { "°" };
        static string[] fmtArray = new string[] { "F0", "F1", "A", "AA", "AAA", "F0AAA"};
        static string[] dir4P = new string[] { "N", "E", "S", "W" };
        static string[] dir8P = new string[] { "N", "NE", "E", "SE", "S", "SW", "W", "NW" };
        static string[] dir16P = new string[] { "N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE", "S", "SSW", "SW", "WSW", "W", "WNW",  "NW", "NNW" };
        static Conversion[] cnvDelegates = new Conversion[] { DtoUnit };

        public Direction()
            : base(unitStr, cnvDelegates)
        {
            m_value = double.NaN;
            m_fmtArray = fmtArray;
        }

        public double Degrees
        {
            get { return m_value; }
            set { m_value = value; }
        }

        public override string Text
        {
            get
            {
                if (char.ToUpper(m_fmtStr[0]) != 'F')
                {
                    switch (m_fmtStr)
                    {
                        case "A":
                            return dir4P[(int)(((m_value + 360+45)%360) / 90)];
                        case "AA":
                            return dir8P[(int)(((m_value + 360 + 22.5) % 360) / 45)];
                        case "AAA":
                            return dir16P[(int)(((m_value + 360 + 11.25) % 360) / 22.5)];
                        case "F0AAA":
                            return (int)m_value + dir16P[(int)(((m_value + 360 + 11.25) % 360) / 22.5)];
                    }
                }

                return base.Text;
            }
        }
    }

    #endregion

    [Serializable]
    class Precipitation : SDistance
    {
        /// <summary>
        /// Precipitation, special case <0 = trace
        /// </summary>
        public Precipitation()
        {
        }

        public const double Trace = -1;
        public override string Text
        {
            get
            {
                return (IsPopulated && m_value < 0) ? "Trace" : base.Text;
            }
        }

        public override string TextandUnit
        {
            get
            {
                return (IsPopulated && m_value < 0) ? Text : base.TextandUnit;
            }
        }
    }

    [Serializable]
    class Humidity : Percent
    {
        public Humidity()
        {
        }
    }

    [Serializable]
    class Visibility : LDistance
    {
        public Visibility()
            : base()
        {
            m_how = How.Exact;
        }

        public enum How { LessThan, Exact, GreaterThan };
        How m_how;

        public How HowTo
        {
            get { return m_how; }
            set { m_how = value; }
        }

        public override string Text
        {
            get
            {
                switch (m_how)
                {
                    case How.LessThan:
                        return "< " + base.Text;
                    case How.GreaterThan:
                        return "> " + base.Text;
                }

                return base.Text;
            }
        }
    }

    [Serializable]
    class WindSpeed : Speed
    {
        public WindSpeed()
            : base()
        {
            m_special = string.Empty;
        }

        string m_special;

        public string Special
        {
            get { return m_special; }
            set { m_special = value; }
        }

        public override string Text
        {
            get
            {
                return (m_special != string.Empty) ? m_special : base.Text;
            }
        }

        public override string TextandUnit
        {
            get
            {
                return (m_special != string.Empty) ? m_special : base.TextandUnit;
            }
        }
    }

    [Serializable]
    class WindDir : Direction
    {
        public WindDir()
            : base()
        {
            m_special = string.Empty;
        }

        string m_special;
        public string Special
        {
            get { return m_special; }
            set { m_special = value; }
        }

        public override string Text
        {
            get { return (m_special.Length != 0)  ? m_special : base.Text; }
        }

        public override string TextandUnit
        {
            get
            {
                return (m_special.Length != 0) ? m_special : base.TextandUnit;
            }
        }
    }

    [Serializable]
    class Wind
    {
        public Wind()
        {
        }

        WindDir m_direction = new WindDir();
        WindDir m_varDirBeg = new WindDir();
        WindDir m_varDirEnd = new WindDir();
        WindSpeed m_speed = new WindSpeed();
        WindSpeed m_gust = new WindSpeed();

        public WindDir Direction
        {
            get { return m_direction; }
            set { m_direction = value; }
        }

        public WindDir VarDirBeg
        {
            get { return m_varDirBeg; }
            set { m_varDirBeg = value; }
        }

        public WindDir VarDirEnd
        {
            get { return m_varDirEnd; }
            set { m_varDirEnd = value; }
        }

        public WindSpeed Speed
        {
            get { return m_speed; }
            set { m_speed = value; }
        }

        public WindSpeed Gust
        {
            get { return m_gust; }
            set { m_gust = value; }
        }
    }

    [Serializable]
    class Runway
    {
        public Runway()
            : base()
        {
        }

        public string Id
        {
            set { }
        }

        public string Approach
        {
            set { }
        }

        public string Tendency
        {
            set { }
        }
        public string MinTendency
        {
            set { }
        }
        public string MaxTendency
        {
            set { }
        }

        public enum How { LessThan, Exact, GreaterThan };
        How m_how;

        public How HowTo
        {
            get { return m_how; }
            set { m_how = value; }
        }

        public LDistance Length = new LDistance();
        public LDistance MinLength = new LDistance();
        public LDistance MaxLength = new LDistance();
    }

    [Serializable]
    class WxTime : WxBase
    {
        public enum Units { Gmt = 0, Local };
        static string[] tmUnitStrs = new string[] { "gmt", "local" };
        static Conversion[] tmCnvDelegates = new Conversion[] { GmtToUnit, LocalToUnit };

        static private double GmtToUnit(double dt, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Gmt:
                    return dt;
                case (int)Units.Local:
                    return dt + TimeZone.CurrentTimeZone.GetUtcOffset(DateTime.FromOADate(dt)).TotalSeconds;
            }

            throw (new Exception("unknown units"));
        }
        static private double LocalToUnit(double dt, int toUnit)
        {
            switch (toUnit)
            {
                case (int)Units.Gmt:
                    return dt - TimeZone.CurrentTimeZone.GetUtcOffset(DateTime.FromOADate(dt)).TotalSeconds;
                case (int)Units.Local:
                    return dt;
            }

            throw (new Exception("unknown units"));
        }

        static string[] defFmtStrs = new string[] {
            "yyyy/MM/dd HH:mm", 
            "dd HH:mm", 
            "HH:mm", 
            "HH:mm:ss", 
            "HH:mm:ss.fff" 
        };

        public WxTime()
            : base(tmUnitStrs, tmCnvDelegates)
        {
            m_fmtArray = defFmtStrs;
            m_fmtStr = m_fmtArray[0];
            m_unit = 0;
        }

        public override bool IsNumber
        {
            get
            {
                return false;
            }
        }
    }

    [Serializable]
    class WxTimeSpan : WxBase
    {
        public enum Units { Default=0 };
        static string[] tmUnitStrs = new string[] { "default" };
        static Conversion[] tmCnvDelegates = new Conversion[] { null };

        static string[] defFmtStrs = new string[] {
            "HH:mm", 
            "HH:mm:ss", 
            "HH:mm:ss.fff" 
        };

        public WxTimeSpan()
            : base(tmUnitStrs, tmCnvDelegates)
        {
            m_fmtArray = defFmtStrs;
            m_fmtStr = m_fmtArray[0];
            m_unit = 0;
        }

        public override string Text
        {
            get
            {
                TimeSpan ts = new TimeSpan((long)this.Value);
#if false
                DateTime dt = new DateTime(2000, 1, 2, 0, 0, 0) + ts;
                return dt.ToString(m_fmtStr);
#else
                string text = m_fmtStr;
                int hours = (int)ts.TotalHours;
                text = text.Replace("HH", hours.ToString("##0"));
                text = text.Replace("mm", ts.Minutes.ToString("00"));
                text = text.Replace("ss", ts.Seconds.ToString("00"));
                text = text.Replace("fff", ts.Milliseconds.ToString("000"));
                return text;
#endif
            }

            // set { base.Text = value; }
        }
    }
}
