// Convert  noaa station list to more compact form.
//
// Semicolon separated list:
//  http://weather.noaa.gov/tg/site.shtml
//
// Fixed position list:
// http://www.rap.ucar.edu/weather/surface/stations.txt
//

// #include "stdafx.h"
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <string>

using namespace std;

double Geopoint(const char* str)
{
    // DDD-MM-SSH   H=N or S or E or W
    double gp = 0;
    char* minPtr;
    char* secPtr;
    char* hemPtr;
    int ddd = strtol(str, &minPtr, 10);
    int mm = strtol(minPtr+1, &secPtr, 10);
    int ss = strtol(secPtr+1, &hemPtr, 10);
    if (hemPtr > secPtr && secPtr > minPtr)
    {
        gp = ddd + (ss/60.0 + mm) / 60.0;
        char h = (*hemPtr != '\0') ? tolower(*hemPtr) : tolower(*secPtr);
        if (h == 'w' || h == 's')
            gp = -gp;
    }
    return gp;
}

// Parse semicolon field separated stations list.

/*
Input;
72;509;KBOS;BOSTON LOGAN INTERNATIONAL;MA;United States;4;42-22N;071-01W;42-21-51N;071-01-09W;6;54;

 0 Block Number	    2 digits representing the WMO-assigned block.
 1 Station Number	3 digits representing the WMO-assigned station.
 2 ICAO Location    4 alphanumeric characters, not all stations in this file have an assigned location indicator. The value "----" is used for stations that do not have an assigned location indicator.
 3 Place Name	      Common name of station location.
 4 State	        2 character abbreviation (included for stations located in the United States only).
 5 Country Name	      Country name is ISO short English form.
 6 WMO Region	       digits 1 through 6 representing the corresponding WMO region, 7 stands for the WMO Antarctic region.
 7 Station Latitude	    DD-MM-SSH where DD is degrees, MM is minutes, SS is seconds and H is N for northern hemisphere or S for southern hemisphere. The seconds value is omitted for those stations where the seconds value is unknown.
 8 Station Longitude    DDD-MM-SSH where DDD is degrees, MM is minutes, SS is seconds and H is E for eastern hemisphere or W for western hemisphere. The seconds value is omitted for those stations where the seconds value is unknown.
 9 Upper Air Latitude	DD-MM-SSH where DD is degrees, MM is minutes, SS is seconds and H is N for northern hemisphere or S for southern hemisphere. The seconds value is omitted for those stations where the seconds value is unknown.
10 Upper Air Longitude	DDD-MM-SSH where DDD is degrees, MM is minutes, SS is seconds and H is E for eastern hemisphere or W for western hemisphere. The seconds value is omitted for those stations where the seconds value is unknown.
11 Station Elevation (Ha)	 The station elevation in meters. Value is omitted if unknown.
12 Upper Air Elevation (Hp)	 The upper air elevation in meters. Value is omitted if unknown.
13 RBSN indicator	P if station is defined by the WMO as belonging to the Regional Basic Synoptic Network, omitted otherwise.


Output:
 
0   WMO(5)
1   ICAO(4)
2   Place Name
3   State
4   Country
5   Latitude
6   Longitude
*/

int StationSemicolonConvert(const char* inStation, const char* outStation)
{
    std::ifstream in(inStation);
    std::ofstream out(outStation);

    if (in.bad())
    {
        std::cout << "Failed to open:" << inStation << std::endl;
        return -1;
    }
    if (out.bad())
    {
        std::cout << "Failed to create:" << outStation << std::endl;
        return -1;
    }

    unsigned long stationCnt = 0;
    char buffer[512];
    std::vector<std::string> fields;

    while (in.getline(buffer, sizeof(buffer)))
    {
        const char sep = ';';
        char* pToken = buffer;
        char* pNext;
        int fieldn = 0; 
        fields.clear();

        while (pNext = strchr(pToken, sep))
        {
            *pNext = '\0';
            fields.push_back(pToken);
            pToken = pNext + 1;
        }

        if (fields.size() > 8 && fields[2].length() > 3 && fields[2][0] != '-' )
        {
            stationCnt++;
            out 
                << fields[0] << fields[1] << sep
                << fields[2] << sep
                << fields[3] << sep
                << fields[4] << sep
                << fields[5] << sep
                << Geopoint(fields[7].c_str()) << sep
                << Geopoint(fields[8].c_str()) << endl;
        }
    }

    std::cout << "Stations converted:" << stationCnt << std::endl;
    return 0;
}

// Parse space position station list.
// !   CD = 2 letter state (province) abbreviation                               
// !   STATION = 16 character stat!ion long name                                  
// !   ICAO = 4-character international id                                       
// !   IATA = 3-character (FAA) id                                               
// !   SYNOP = 5-digit international synoptic number                             
// !   LAT = Latitude (degrees minutes)                                          
// !   LON = Longitude (degree minutes)                                          
// !   ELEV = Station elevation (meters)                                         
// !   M = METAR reporting station.   Also Z=obsolete? site                      
// !   N = NEXRAD (WSR-88D) Radar site                                           
// !   V = Aviation-specific flag (V=AIRMET/SIGMET end point, A=ARTCC T=TAF U=T+V)
// !   U = Upper air (rawinsonde=X) or Wind Profiler (W) site                    
// !   A = Auto (A=ASOS, W=AWOS, M=Meso, H=Human, G=Augmented) (H/G not yet impl.)
// !   C = Office type F=WFO/R=RFC/C=NCEP Center                                 
// !   Digit that follows is a priority for plotting (0=highest)
// !   Country code (2-char) is last column
// !
//           1         2         3         4         5         6         7         8
// 0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
// !                                                                             
// 
// ALASKA             16-DEC-08                                                  
// CD  STATION         ICAO  IATA  SYNOP   LAT     LONG   ELEV   M  N  V  U  A  C
// AK ADAK NAS         PADK  ADK   70454  51 53N  176 39W    4   X     T          7 US
// AK AKHIOK           PAKH  AKK          56 56N  154 11W   14   X                8 US
// 0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//           1         2         3         4         5         6         7         8
//
// Output:
// 
// 0   WMO(5)
// 1   ICAO(4)
// 2   Place Name
// 3   State
// 4   Country
// 5   Latitude
// 6   Longitude
//
int StationPosConvert(const char* inStation, const char* outStation, const char* inFilter)
{
    std::map<std::string, bool> filter;
    if (inFilter != NULL)
    {
        std::ifstream inf(inFilter);
        char fbuffer[512];
        while (inf.getline(fbuffer, sizeof(fbuffer)))
        {
            strcat(fbuffer, "     ");
            char icao[10];
            sprintf(icao, "%5.5s", fbuffer);      // force to 5 characters.
            filter.insert(make_pair(icao, 1));
        }
    }

    std::ifstream in(inStation);
    std::ofstream out(outStation);

    if (in.bad())
    {
        std::cout << "Failed to open:" << inStation << std::endl;
        return -1;
    }
    if (out.bad())
    {
        std::cout << "Failed to create:" << outStation << std::endl;
        return -1;
    }

    char buffer[512];
    int begPos[] = {0,  3, 20, 26, 32, 39, 47, 56, 62, 81}; 
    int endPos[] = {2, 19, 25, 31, 38, 45, 54, 61, 80, 83}; 
    int cntPos = sizeof(endPos) / sizeof(endPos[0]);
    unsigned long stationCnt = 0;

// #define TEST
    char* sTest = "   SYDNEY INTL AIRP YSSY        94767  33 57S  151 10E    3   X     T          6 AU";
 
#ifdef TEST
    strcpy(buffer, sTest);
#else
    while (in.getline(buffer, sizeof(buffer)))
#endif
    {
        int bufLen = strlen(buffer);

#ifndef TEST
        if (bufLen < 81)
            continue;
#endif

        bool isGood = (buffer[0] != '!');
        for (int i = 0; i < cntPos && isGood; i++)
        {
            isGood = (buffer[endPos[i]] == ' ' || buffer[endPos[i]] == '\0');
            buffer[endPos[i]] = '\0';
        }

        if (isGood && strcspn(buffer + begPos[2], " ") > 2)
        {
            if (inFilter == NULL || filter.find(buffer + begPos[2]) != filter.end())
            {
                const char sep = ';';
                out 
                    << buffer + begPos[4] << sep                    // wmo
                    << buffer + begPos[2] << sep                    // icao
                    << buffer + begPos[1] << sep                    // place
                    << buffer + begPos[0] << sep                    // state
                    << buffer + begPos[9] << sep                    // country
                    << Geopoint(buffer + begPos[5]) << sep          // latitude
                    << Geopoint(buffer + begPos[6]) << endl;        // longitude

                stationCnt++;
            }
        }
    }

    std::cout << "Stations converted:" << stationCnt << std::endl;
    return 0;
}


int main(int argc,char* argv[])
{
    if (argc == 4 || argc == 5)
    {
        switch (tolower(argv[1][0]))
        {
        case 's':
        case ';':
            return StationSemicolonConvert(argv[2], argv[3]);
        case 'p':
            return StationPosConvert(argv[2], argv[3], (argc == 5) ? argv[4] : NULL);
        default:
            std::cout << "Unknown station type, should be semi or pos\n";
        }
    }
  

    std::cout << "StationConvert <semi|pos> <inputSemicolonStationList> <outputStation>\n";
    std::cout << "Semicolon separated list can be found at:\n"
        "   http://weather.noaa.gov/tg/site.shtml \n"
        "Fix position list can be found at:\n"
        "   http://www.rap.ucar.edu/weather/surface/stations.txt \n"
        "\n"
        "Use optional filterIcao.txt to prune output list to just those Icao in the filter list.\n";

    std::cout << "Examples\n"
        "  StationConvert semi semicolonList.txt mststatn.dat\n" 
        "  StationConvert pos posList.txt mststatn.dat\n"
        "  StationConvert pos posList.txt mststatn.dat filterIcao.txt\n";


	return 0;
}

